import re
import subprocess
from fsbc.system import windows, macosx

if windows:
    # noinspection PyUnresolvedReferences
    try:
        import win32api
    except ImportError:
        win32api = None
    # noinspection PyUnresolvedReferences
    try:
        import win32con
    except ImportError:
        win32con = None
    EDS_RAWMODE = 2
elif macosx:
    # noinspection PyUnresolvedReferences
    import Quartz


class RefreshRateTool(object):

    def __init__(self, game_platform=None, game_refresh=None):
        self.game_platform = game_platform
        self.game_refresh = game_refresh

    def set_best_mode(self):
        print("RefreshRateTool.set_best_mode")
        current = self.get_current_mode()
        print("REFRESH RATE CONTROL: current mode is", current)
        modes = self.get_all_modes()
        mode_score_list = []
        for mode in modes:
            # print(mode)
            # for now, only consider same resolution
            if mode['width'] != current['width']:
                continue
            if mode['height'] != current['height']:
                continue
            # only consider modes with the same bpp
            if mode['bpp'] != current['bpp']:
                continue
            score = self.calculate_mode_score(mode)
            if score is not None:
                mode_score_list.append((score, mode))
        if len(mode_score_list) == 0:
            print("REFRESH RATE TOOL: mode_score_list is empty")
            return
        print("REFRESH RATE TOOL:", mode_score_list)
        best_mode = mode_score_list[0][1]
        if best_mode == current:
            print("REFRESH RATE TOOL: using current mode")
        else:
            self.set_mode(best_mode)

    def calculate_mode_score(self, mode, debug=False):
        game_refresh = self.game_refresh
        if not self.game_refresh:
            if debug:
                print("REFRESH RATE TOOL: game refresh not specified")
            return None
        # in case of values such as -1 etc
        if game_refresh < 1:
            if debug:
                print("REFRESH RATE TOOL: game refresh not specified")
            return None
        display_refresh = mode['refresh']
        if not display_refresh:
            if debug:
                print("REFRESH RATE TOOL: display refresh not found")
            return None
        # in case of values such as -1 etc
        if display_refresh < 1:
            if debug:
                print("REFRESH RATE TOOL: display refresh not found")
            return None
        diff = game_refresh - display_refresh
        # FIXME: configurable?
        # FIXME: configurable per platform?
        allowable_pos_diff = 1.01
        allowable_neg_diff = 1.01
        if debug:
            print("REFRESH RATE TOOL: game refresh: {0} vs "
                  "display refresh: {1}".format(game_refresh, display_refresh))
            print("REFRESH RATE TOOL: allow +{0}/-{1}".format(
                allowable_pos_diff, allowable_neg_diff))
            print("REFRESH RATE TOOL: diff: {0}".format(diff))
        if allowable_pos_diff > diff > -allowable_neg_diff:
            if debug:
                print("REFRESH RATE TOOL: allow vsync")
            return diff
        if debug:
            print("REFRESH RATE TOOL: deny vsync")
        return None

    def allow_vsync(self):
        """
        game_refresh = self.game_refresh
        if not self.game_refresh:
            print("REFRESH RATE TOOL: game refresh not specified")
            return False
        # in case of values such as -1 etc
        if game_refresh < 1:
            print("REFRESH RATE TOOL: game refresh not specified")
            return False
        display_refresh = self.get_display_refresh()
        if not display_refresh:
            print("REFRESH RATE TOOL: display refresh not found")
            return False
        # in case of values such as -1 etc
        if display_refresh < 1:
            print("REFRESH RATE TOOL: display refresh not found")
            return False
        diff = game_refresh - display_refresh
        # FIXME: configurable?
        # FIXME: configurable per platform?
        allowable_pos_diff = 1.01
        allowable_neg_diff = 1.01
        print("REFRESH RATE TOOL: game refresh: {0} vs "
                "display refresh: {1}".format(game_refresh,
                display_refresh))
        print("REFRESH RATE TOOL: allow +{0}/-{1}".format(allowable_pos_diff,
                allowable_neg_diff))
        print("REFRESH RATE TOOL: diff: {0}".format(diff))
        if diff < allowable_pos_diff and diff > -allowable_neg_diff:
            print("REFRESH RATE TOOL: allow vsync")
            return True
        print("REFRESH RATE TOOL: deny vsync")
        """
        current = self.get_current_mode()
        return self.calculate_mode_score(current, debug=True) is not None
        
    def get_display_refresh(self):
        # # FIXME:
        # if macosx:
        #     return 60.0
        # elif windows:
        #     import win32api
        #     device = win32api.EnumDisplayDevices()
        #     settings = win32api.EnumDisplaySettings(device.DeviceName, 0)
        #     return settings.DisplayFrequency()
        # # FIXME:
        # return 50.0
        # return None
        return self.get_current_mode()["refresh"]

    def get_current_mode(self):
        refresh = 0
        width = 640
        height = 480
        bpp = 0
        flags = 0

        if windows:
            if win32api:
                settings = win32api.EnumDisplaySettings(
                    None, win32con.ENUM_CURRENT_SETTINGS)
                refresh = float(settings.DisplayFrequency)
                width = int(settings.PelsWidth)
                height = int(settings.PelsHeight)
                bpp = int(settings.BitsPerPel)
                flags = int(settings.DisplayFlags)
        elif macosx:
            main_display = Quartz.CGMainDisplayID()
            current_display = Quartz.CGDisplayPrimaryDisplay(
                main_display)
            current_mode = Quartz.CGDisplayCopyDisplayMode(current_display)
            width = Quartz.CGDisplayModeGetWidth(current_mode)
            height = Quartz.CGDisplayModeGetHeight(current_mode)
            refresh = Quartz.CGDisplayModeGetRefreshRate(current_mode)
            if not refresh:
                print("WARNING: returned refresh rate was 0. assuming 60 Hz")
                refresh = 60
            # A bit weird that it crashes, since this is supposed to be a
            # copy of the mode...?
            print("FIXME: Not calling Quartz.CGDisplayModeRelease("
                  "current_mode), seems to crash pygame on exit...")
            # Quartz.CGDisplayModeRelease(current_mode)
            flags = 0
            bpp = None
        else:
            return self._get_current_mode_x()

        return {"width": width, "height": height, "refresh": refresh,
                "bpp": bpp, "flags": flags}

    def get_all_modes(self):
        modes = []
        if windows:
            k = 0
            while True:
                try:
                    settings = win32api.EnumDisplaySettingsEx(
                        None, k, EDS_RAWMODE)
                except win32api.error:
                    break
                refresh = float(settings.DisplayFrequency)
                width = int(settings.PelsWidth)
                height = int(settings.PelsHeight)
                bpp = int(settings.BitsPerPel)
                flags = int(settings.DisplayFlags)
                # print(width, height, refresh, bpp, flags)
                modes.append({'width': width, 'height': height,
                              'refresh': refresh, 'bpp': bpp, 'flags': flags})
                k += 1
        elif macosx:
            # FIXME:
            pass
        else:
            modes = self._get_all_modes_x()
        # modes.extend(self.get_override_modes())
        return modes

    # def get_override_modes(self):
    #     modes = []
    #     k = 0
    #     while True:
    #         mode_string = pyapp.user.ini.get('ForceDisplayModes/%d' % k)
    #         if not mode_string:
    #             break
    #         mode_string = mode_string.lower()
    #         mode = {}
    #         w, rest = mode_string.split('x')
    #         h, rest = rest.split('@')
    #         r, rest = rest.split('hz')
    #         mode['width'] = int(w)
    #         mode['height'] = int(h)
    #         mode['bpp'] = 32
    #         mode['refresh'] = int(r)
    #         mode['flags'] = 0
    #         modes.append(mode)
    #         k += 1
    #     print("get_override_modes returning", modes)
    #     return modes

    def set_mode(self, mode):
        # if windows:
        #     self._set_mode_windows(mode)
        # elif macosx:
        #     # FIXME:
        #     print("WARNING: mode settings is not supported on this "
        #             "platform yet")
        # else:
        #     self._set_mode_x(mode)

        print("FIXME: Currently disabled set_mode")
        return

    def _set_mode_windows(self, mode):
        print("REFRESH RATE TOOL: setting mode", mode)
        if not win32api:
            print("win32api not available, returning")
            return False
        k = 0
        while True:
            try:
                settings = win32api.EnumDisplaySettingsEx(None, k, EDS_RAWMODE)
            except win32api.error:
                break
            refresh = float(settings.DisplayFrequency)
            width = int(settings.PelsWidth)
            height = int(settings.PelsHeight)
            bpp = int(settings.BitsPerPel)
            flags = int(settings.DisplayFlags)
            # print(width, height, refresh, bpp, flags)
            # modes.append({'width': width, 'height': height,
            #        'refresh': refresh, 'bpp': bpp, 'flags': flags})
            if width == mode['width'] and \
                    height == mode['height'] and \
                    refresh == mode['refresh'] and \
                    bpp == mode['bpp'] and \
                    flags == mode['flags']:

                # print("trying to override with refresh", int(round(self.game_refresh)))
                # #refresh == mode['refresh'] and \
                # settings.DisplayFrequency = int(round(self.game_refresh))
                # result = win32api.ChangeDisplaySettings(settings,
                #         win32con.CDS_UPDATEREGISTRY) #win32con.CDS_FULLSCREEN)
                #         #0) #win32con.CDS_FULLSCREEN)
                # if result == win32con.DISP_CHANGE_SUCCESSFUL:
                #     print("display change was successful")
                #     return True
                # print("failed, falling back to ", mode['refresh'])
                # settings.DisplayFrequency = mode['refresh']
                print("found windows mode, changing display settings")
                result = win32api.ChangeDisplaySettings(
                    settings, win32con.CDS_UPDATEREGISTRY)
                # win32con.CDS_FULLSCREEN)

                if result == win32con.DISP_CHANGE_SUCCESSFUL:
                    print("display change was successful")
                    return True
                else:
                    print("display change failed, result =", result)
                    return False
            k += 1

        # for omode in self.get_override_modes():
        #     print("trying override mode", omode)
        #     if omode['width'] == mode['width'] and \
        #             omode['height'] == mode['height'] and \
        #             omode['refresh'] == mode['refresh'] and \
        #             omode['bpp'] == mode['bpp'] and \
        #             omode['flags'] == mode['flags']:
        #         settings.PelsWidth = omode['width']
        #         settings.PelsHeight = omode['height']
        #         settings.BitsPerPel = omode['bpp']
        #         settings.DisplayFlags = omode['flags']
        #         settings.DisplayFrequency = omode['refresh']
        #         result = win32api.ChangeDisplaySettings(settings,
        #                 win32con.CDS_UPDATEREGISTRY) #win32con.CDS_FULLSCREEN)
        #                 #0) #win32con.CDS_FULLSCREEN)
        #         if result == win32con.DISP_CHANGE_SUCCESSFUL:
        #             print("display change was successful")
        #             return True
        #         else:
        #             print("display change failed, result =", result)
        #             return False
        return False

    def _set_mode_x(self, mode):
        print("REFRESH RATE TOOL: setting mode", mode)
        args = ['/usr/bin/env', 'xrandr', '-s', '{0}x{1}'.format(
                mode['width'], mode['height']), '-r',
                str(mode['refresh'])]
        p = subprocess.Popen(args)
        p.wait()

    def _get_all_modes_x(self):
        modes = []
        self._get_current_mode_x(modes=modes)
        return modes

    def _get_current_mode_x(self, modes=[]):
        modes[:] = []
        mode = {'width': 0, 'height': 0, 'refresh': 0.0, 'bpp': 0,
                'flags': 0} 
        args = ['/usr/bin/env', 'xrandr', '-q']
        p = subprocess.Popen(args, stdout=subprocess.PIPE)
        for line in p.stdout:
            line = line.decode("ISO-8859-1")
            if not line.startswith(' '):
                continue
            line = line.strip()
            line = line.replace("+", "")
            # collapse multiple spaces
            line = re.sub(' +', ' ', line)
            parts = line.split(' ')
            resolution = parts[0]
            rates = parts[1:]
            width, height = resolution.split('x')
            width = int(width)
            height = int(height)
            for rate in rates:
                if rate[-1] == '*':
                    refresh = float(rate[:-1])
                    mode['width'] = width
                    mode['height'] = height
                    mode['refresh'] = refresh
                    modes.append({'width': width, 'height': height,
                                  'refresh': float(rate[:-1]), 'bpp': 0,
                                  'flags': 0})
                else:
                    modes.append({'width': width, 'height': height,
                                  'refresh': float(rate), 'bpp': 0,
                                  'flags': 0})
        return mode



# this is a quick hack tp fix problems with some modules not being imported
# by py2exe / py2app otherwise

import fs_uae_workspace.apps.adf_creator_app
import fs_uae_workspace.apps.clock_app
import fs_uae_workspace.apps.hdf_creator_app
import fs_uae_workspace.apps.joystick_config_app
import fs_uae_workspace.apps.launcher_app
import fs_uae_workspace.apps.locker_uploader
import fs_uae_workspace.apps.login
import fs_uae_workspace.apps.logout
import fs_uae_workspace.apps.refresh

import fs_uae_workspace.prefs.audio
import fs_uae_workspace.prefs.experimental_features
import fs_uae_workspace.prefs.filter
import fs_uae_workspace.prefs.game_database
import fs_uae_workspace.prefs.joystick
import fs_uae_workspace.prefs.keyboard
import fs_uae_workspace.prefs.language
import fs_uae_workspace.prefs.mouse
import fs_uae_workspace.prefs.netplay
# import fs_uae_workspace.prefs.opengl
import fs_uae_workspace.prefs.scan
import fs_uae_workspace.prefs.video



from .ui.LauncherFilePicker import LauncherFilePicker
from fsgs.amiga.Amiga import Amiga
from .Config import Config
from .I18N import gettext
from fsbc.Paths import Paths
from fsgs.FSGSDirectories import FSGSDirectories


class CDManager(object):

    @classmethod
    def clear_all(cls):
        for i in range(1):
            cls.eject(i)
        cls.clear_cdrom_list()

    @classmethod
    def eject(cls, drive):
        values = [("cdrom_drive_{0}".format(drive), ""),
                  ("x_cdrom_drive_{0}_sha1".format(drive), "")]
        Config.set_multiple(values)

    @classmethod
    def clear_cdrom_list(cls):
        values = []
        for i in range(Amiga.MAX_CDROM_IMAGES):
            values.append(("cdrom_image_{0}".format(i), ""))
            values.append(("x_cdrom_image_{0}_sha1".format(i), ""))
        Config.set_multiple(values)

    @classmethod
    def multiselect(cls, parent=None):
        default_dir = FSGSDirectories.get_cdroms_dir()
        dialog = LauncherFilePicker(
            parent, gettext("Select Multiple CD-ROMs"), "cd", multiple=True)

        if not dialog.show_modal():
            return
        paths = dialog.get_paths()
        paths.sort()

        # checksum_tool = ChecksumTool(parent)
        for i, path in enumerate(paths):
            # sha1 = checksum_tool.checksum(path)
            sha1 = ""
            print("FIXME: not calculating CD checksum just yet")
            path = Paths.contract_path(path, default_dir)

            if i < 1:
                Config.set_multiple([
                    ("cdrom_drive_{0}".format(i), path),
                    ("x_cdrom_drive_{0}_sha1".format(i), sha1)
                ])
            Config.set_multiple([
                ("cdrom_image_{0}".format(i), path),
                ("x_cdrom_image_{0}_sha1".format(i), sha1)
            ])

        # blank the rest of the drives
        for i in range(len(paths), 1):
            Config.set_multiple([
                ("cdrom_drive_{0}".format(i), ""),
                ("x_cdrom_drive_{0}_sha1".format(i), "")
            ])

            # Config.set("x_cdrom_drive_{0}_sha1".format(i), "")
            # Config.set("x_cdrom_drive_{0}_name".format(i), "")
        # blank the rest of the image list
        for i in range(len(paths), Amiga.MAX_CDROM_IMAGES):
            Config.set_multiple([
                ("cdrom_image_{0}".format(i), ""),
                ("x_cdrom_image_{0}_sha1".format(i), "")
            ])



from fsui.qt import QDesktopWidget, QMainWindow
from .x11 import configure_dock_window


class TitleBarWindow(QMainWindow):

    def __init__(self):
        super().__init__()
        self.setWindowTitle("FS-UAE Workspace Title Bar")
        g = QDesktopWidget().screenGeometry()
        self.setFixedSize(g.width(), 28)
        self.move(0, 0)
        configure_dock_window(self, 0, 0, 28, 0)



import fsui


class PreferencesHeader(fsui.Group):

    def __init__(self, parent, icon, title, subtitle):
        fsui.Group.__init__(self, parent)
        self.layout = fsui.HorizontalLayout()
        # self.layout.set_padding(0, 20, 0, 20)

        heading_layout = self.layout
        heading_layout.add(fsui.ImageView(self, icon), fill=-1)
        heading_layout.add_spacer(20)
        heading_layout_2 = fsui.VerticalLayout()
        heading_layout.add(
            heading_layout_2, expand=True, fill=False, valign=0.5)
        heading_layout_2.add(fsui.HeadingLabel(self, title))
        heading_layout_2.add_spacer(4)
        heading_layout_2.add(fsui.Label(self, subtitle))



from fsgs.context import fsgs
from game_center.option import Option
from fsgs.platform import PlatformHandler
from game_center.glui.navigatable import Navigatable
from game_center.glui.opengl import *
from game_center.glui.font import BitmapFont
from game_center.glui.menu import Menu
from game_center.glui.items import GameCenterItem, HomeItem, MenuItem
from game_center.glui.state import State
from game_center.glui.render import Render
from game_center.glui.texture import Texture
from game_center.glui.inputmenu import InputMenu


# FIXME: REMOVE?
class Transition(object):
    value = 0.0
    start = 0
    end = 0
    on_finish = None


def render_wall():  # brightness=1.0):
    # glClearColor(0.1, 0.1, 0.1, 1.0)
    # glClear(GL_DEPTH_BUFFER_BIT);
    Render.hd_perspective()
    # fs_emu_ortho();
    # fs_emu_blending(FALSE);
    # fs_emu_texturing(FALSE);
    z = -0.999
    
    # glPushMatrix()
    # glTranslate(0.0, 0.0, 1.0)
    
    # transition y-coordinate between floor and wall
    splt = 361
    fs_emu_blending(False)
    fs_emu_texturing(False)
    glBegin(GL_QUADS)
    
    glColor3f(39.0 / 255.0, 44.0 / 255.0, 51.0 / 255.0)
    glVertex3f(0, splt, z)
    glVertex3f(1920, splt, z)
    color = 0
    glColor3f(color, color, color)
    glVertex3f(1920, 1020, z)
    glVertex3f(0, 1020, z)

    glVertex3f(0, 1020, z)
    glVertex3f(1920, 1020, z)
    glVertex3f(1920, 1080, z)
    glVertex3f(0, 1080, z)
    
    color = 0
    glColor3f(color, color, color)
    glVertex3f(0, 0, z)
    glVertex3f(1920, 0, z)
    glColor3f(20.0 / 255.0, 22.0 / 255.0, 26.0 / 255.0)
    glVertex3f(1920, splt, z)
    glVertex3f(0, splt, z)
    
    glEnd()
    # fs_emu_texturing(True)

    # fs_emu_perspective();
    # double t_x = t0_x + (t1_x - t0_x) * g_menu_transition;
    # double t_y = t0_y + (t1_y - t0_y) * g_menu_transition;
    # double t_z = t0_z + (t1_z - t0_z) * g_menu_transition;
    # double r_a = r0_a + (r1_a - r0_a) * g_menu_transition;
    #
    # glScaled(16.0 / 9.0, 1.0, 1.0);
    # glTranslated(t_x, t_y, t_z);
    # glRotated(r_a, 0.0, 1.0, 0.0);


class GameMenu(Menu):

    def __init__(self, item):
        Menu.__init__(self)

        self.runner = None
        self.platform_handler = None
        self.last_menu_data = None

        self.temp_fix_configs(item)

        self.items.append(item)
        if self.use_game_center_item():
            self.top.left.append(GameCenterItem())
        self.top.left.append(HomeItem())
        self.top.left.append(MenuItem(item.title))
        
        self.context = None
        self.controller = None

        self.create_context()
        self.create_controller()
        print("GameMenu.__init__: controller is", self.controller)

        self.config_list = GameConfigList(self.runner, item)
        self.info_panel = GameInfoPanel()

        self.config_list.info_panel = self.info_panel
        # self.config_list.controller = self.controller
        self.info_panel.config_list = self.config_list

        self.navigatable = self.config_list

        # reset transition
        Transition.start = 0

    def temp_fix_configs(self, item):
        from fsgs.Database import Database
        local_game_database = Database.get_instance()
        game_database = fsgs.get_game_database()

        variants = local_game_database.find_game_variants_new(
            game_uuid=item.uuid)
        print(variants)

        ordered_list = []
        for variant in variants:

            variant["like_rating"], variant["work_rating"] = \
                game_database.get_ratings_for_game(variant["uuid"])
            variant["personal_rating"], ignored = \
                local_game_database.get_ratings_for_game(variant["uuid"])

            # user_rating = variant[5] or 0
            # global_rating = variant[3] or 0
            # user_rating = 0
            # global_rating = 0

            variant_uuid = variant["uuid"]  # variant[2]
            variant_name = variant["name"]  # variant[1]
            variant_name = variant_name.replace("\n", " (")
            variant_name = variant_name.replace(" \u00b7 ", ", ")
            variant_name += ")"
            ordered_list.append(
                ((1 - bool(variant["have"]),
                  1000 - variant["personal_rating"],
                  1000 - variant["like_rating"]),
                 (variant_uuid, variant_name, variant["database"])))
        ordered_list.sort()
        print("ordered variant list:")
        for variant in ordered_list:
            print("-", variant[1][1])
        item.configurations = [co[1] for co in ordered_list]

    def create_context(self):
        item = self.items[0]
        variant_uuid = item.configurations[0][0]
        database_name = item.configurations[0][2]
        print("\nitem:\n", item.configurations)
        print("\nitem[0]:\n", item.configurations[0])
        print("\n\nvariant_uuid =", variant_uuid, "\n\n")

        # print("\n\ncreate_context")
        # print("configurations: ", item.configurations)
        # configs = sort_configurations(item.configurations)

        # configs = item.configurations
        # config = configs[0]
        # self.context = GameContext.create_for_game(
        #     item.platform, item.name, config)

        values = fsgs.game.set_from_variant_uuid(database_name, variant_uuid)

        # print("")
        # for key in sorted(values.keys()):
        #     print(" * {0} = {1}".format(key, values[key]))
        # print("")

        self.platform_handler = PlatformHandler.create(fsgs.game.platform.id)
        loader = self.platform_handler.get_loader(fsgs)
        fsgs.config.load(loader.load_values(values))

    def create_controller(self):
        # old_controller = self.controller
        # self.controller = GameHandler.create_for_game_context(
        #     self.context)

        self.runner = self.platform_handler.get_runner(fsgs)

        # self.controller = GameRunner()

    def recreate_controller(self):
        # self.create_controller()
        # print("***** recreate_controller, new controller is",
        #         id(self.controlleGame Controlr))
        # self.config_list.set_controller(self.controller)
        print("recreate_controller, actually, just recreating game context")
        self.create_context()
        self.controller.context = self.context

    def render(self):
        Transition.value = 0.0
        if Transition.start > 0:
            Transition.value = (State.time - Transition.start) / \
                               (Transition.end - Transition.start)
            # prevent render from stopping when animating
            Render.dirty = True
        # transition goes from 1.0 ... 0.0
        Transition.value = 1.0 - Transition.value
        # finished = 0
        if Transition.value <= 0.0:
            Transition.value = 0.0
            # finished = 1.0
            Transition.start = 0
            Transition.on_finish()
            return
        
        last_menu = State.history[-2]
        self.last_menu_data = last_menu.render()
        
        # self.top_menu_transition = transition
        
        # glClear(GL_DEPTH_BUFFER_BIT)
        # print(transition)
        # glEnable(GL_DEPTH_TEST)
        # glDepthMask(True)
        
        # render_wall(transition)
        # render_screen(transition)
        # if finished:
        #     # FIXME: defer to later via an event or similar?
        return None

    def render_transparent(self, data):
        last_menu = State.history[-2]
        last_menu.render_transparent(self.last_menu_data)
        
        if Transition.value < 1.0:
            opacity = 1.0 - Transition.value
            opacity *= 1.25
            if opacity > 1.0:
                opacity = 1.0
            Render.hd_perspective()
            glDisable(GL_DEPTH_TEST)
            fs_emu_texturing(False)
            fs_emu_blending(True)
            glBegin(GL_QUADS)
            glColor4f(0.0, 0.0, 0.0, opacity)
            glVertex2f(0.0, 0.0)
            glVertex2f(1920.0, 0.0)
            glVertex2f(1920.0, 1020.0)
            glVertex2f(0.0, 1020.0)
            glEnd()
            glEnable(GL_DEPTH_TEST)
        
        self.config_list.render(Transition.value)

    def go_up(self):
        print("GameMenu.go_up")

    def go_down(self):
        print("GameMenu.go_down")

    def go_left(self, count=1):
        print("GameMenu.go_left")

    def go_right(self, count=1):
        print("GameMenu.go_right")

    def activate(self):
        print("GameMenu.activate")


class GameConfigList(Navigatable):

    def __init__(self, controller, game_item):
        self.index = 0
        self.runner = None
        self.game_item = game_item
        self.items = []
        self.set_controller(controller)

    def set_controller(self, controller):

        self.items = []
        self.items.append(Option.create_group("Game Control", -2.5))
        play_option = Option.create_play_option()
        self.items.append(play_option)

        # self.items = controller.options
        self.index = 0
        # set index to first non-group item
        for i, item in enumerate(self.items):
            if not item.group:
                self.index = i
                break
        self.runner = controller
        # self.controller = controller
        # print("GameConfigList.set_controller", id(self.controller))

    def go_up(self):
        old_index = self.index
        if self.index > 0:
            new_index = self.index - 1
            while new_index > 0 and self.items[new_index].group:
                new_index -= 1
        else:
            new_index = 0
        if new_index != old_index and not self.items[new_index].group:
            self.index = new_index
        else:
            Navigatable.go_up(self)

    def go_down(self):
        old_index = self.index
        if self.index < len(self.items) - 1:
            new_index = self.index + 1
            while new_index < len(self.items) - 1 and \
                    self.items[new_index].group:
                new_index += 1
        else:
            new_index = len(self.items) - 1
        if new_index != old_index and not self.items[new_index].group:
            self.index = new_index

    def go_left(self, count=1):
        print("GameConfigList.go_left")
        pass

    def go_right(self, count=1):
        pass

    def activate(self):
        result = self.items[self.index].activate()
        if result == "PLAY":

            # noinspection PyDecorator
            @staticmethod
            def show_input():
                # print("Create input menu, controller = ", id(self.controller))
                # new_menu = InputMenu(self.game_item, self.controller)
                # State.history.append(new_menu)
                # # FIXME
                # from .window import set_current_menu
                # set_current_menu(new_menu)

                print("Create input menu, runner = ", id(self.runner))
                new_menu = InputMenu(self.game_item, self.runner)
                State.history.append(new_menu)
                # FIXME
                from game_center.glui.window import set_current_menu
                set_current_menu(new_menu)

            Transition.start = State.time
            Transition.end = Transition.start + 0.4
            Transition.on_finish = show_input

    def render(self, transition=1.0):
        Render.hd_perspective()
        
        # w = 560
        # h = 92
        # x = 1920 - w * transition
        # y = 1080 - 60 - h
        # y = 1080 - 60 - h - 44
        # z = -0.9
        
        item_top = 1080 - 90
        item_left = 1920 - 560 + 40

        item_left += Texture.sidebar_background_shadow.w * (
            1.0 - transition)
        
        glPushMatrix()
        glTranslate(0.0, 0.0, 0.7)
        # glTranslated((1.0 - transition) * g_tex_sidebar_background->width, 0, 0)
        # fs_emu_set_texture(g_tex_sidebar_background)
        # fs_emu_render_texture_with_size(g_tex_sidebar_background,
        #         1920 - g_tex_sidebar_background->width,
        #         0, g_tex_sidebar_background->width,
        #         g_tex_sidebar_background->height)
        # fs_emu_render_sidebar()
        fs_emu_blending(True)
        fs_emu_texturing(True)
        glDepthMask(False)
        Texture.sidebar_background_shadow.render(
            1920 - Texture.sidebar_background_shadow.w * transition,
            0, Texture.sidebar_background_shadow.w,
            Texture.sidebar_background_shadow.h)
        glDepthMask(True)
        glPopMatrix()
        
        # x = x + 70
        # w = w - 70
        x = item_left
        y = item_top - 60
        w = 560 - 40 - 40
        h = 60
        
        for i, item in enumerate(self.items):
            if item.group:
                if i > 0:
                    # skip one half row before group heading
                    y -= h / 2
            selected = i == self.index and State.navigatable == self
            z = 0.71
            fs_emu_texturing(False)
            
            if selected:
                fg_color = [1.0, 1.0, 1.0, 1.0]
                glBegin(GL_QUADS)

                glColor3f(0.00, 0x99 / 0xff, 0xcc / 0xff)
                glVertex3f(x, y + 4, z)
                glVertex3f(x + w, y + 4, z)
                glVertex3f(x + w, y + h - 4, z)
                glVertex3f(x, y + h - 4, z)
                # glColor3f(0.6, 0.6, 0.6)
                # glVertex3f(x, y + 4, z)
                # glVertex3f(x + w, y + 4, z)
                # glVertex3f(x + w, y + h, z)
                # glVertex3f(x, y + h, z)
                glEnd()
            else:
                fg_color = [1.0, 1.0, 1.0, 1.0]

            # glBegin(GL_QUADS)
            # glColor3f(0.4, 0.4, 0.4)
            # glVertex3f(x, y, z)
            # glVertex3f(x + w, y, z)
            # glVertex3f(x + w, y + 4, z)
            # glVertex3f(x, y + 4, z)
            # glEnd()
            # fs_emu_texturing(True)

            if i > 0:
                fg_color[3] *= 0.35

            text = item.title
            
            # FIXME: REMOVE
            # fs_emu_blending(False)
            # fs_emu_blending(True)
            
            glDisable(GL_DEPTH_TEST)            
            if item.group:
                BitmapFont.menu_font.render(
                    text, x + 20, y + 14, r=0.0, g=0x99 / 0xff, b=0xcc / 0xff)
            else:
                BitmapFont.menu_font.render(text, x + 20, y + 14)
            glEnable(GL_DEPTH_TEST)
            # text = item.title.upper()
            # tw, th = Render.text(text, Font.main_path_font,
            #         x + 40, y + 43, color=fg_color)
            # text = item.subtitle.upper()
            # fg_color[3] = fg_color[3] * 0.4
            # text = item.title.upper()
            # tw, th = Render.text(text, Font.list_subtitle_font,
            #         x + 40, y + 18, color=fg_color)
            y -= h


class GameInfoPanel(Navigatable):

    def __init__(self):
        self.index = 0
        self.position = 0

    def go_up(self):
        print("GameInfoPanel.go_up")
        if self.index == 0 and self.position == 0:
            Navigatable.go_up(self)

    def go_down(self):
        print("GameInfoPanel.go_down")

    def go_left(self, count=1):
        print("GameInfoPanel.go_left")

    def go_right(self, count=1):
        State.navigatable = self.config_list

    def render(self):
        z = -0.999
        fs_emu_texturing(False)
        fs_emu_blending(False)
        glBegin(GL_QUADS)
        glColor3f(0.0, 0.0, 0.0)
        # glVertex3f(0, 365, z)
        # glVertex3f(1270, 365, z)
        glVertex3f(0, 400, z)
        glVertex3f(1270, 400, z)
        glColor3f(0.25, 0.25, 0.25)
        glVertex3f(1270, 1080 - 60, z)
        glVertex3f(0, 1080 - 60, z)
        glEnd()
        # fs_emu_texturing(True)



import fsui as fsui
from ...I18N import gettext
from .OptionUI import OptionUI
from fsui.extra.iconheader import NewIconHeader


class KeyboardSettingsPage(fsui.Panel):

    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        self.layout = fsui.VerticalLayout()
        # self.layout.set_padding(20, 20, 20, 20)

        self.icon_header = NewIconHeader(
            self, fsui.Icon("keyboard-settings", "pkg:fs_uae_workspace"),
            gettext("Keyboard Settings"),
            "")
        self.layout.add(self.icon_header, fill=True, margin_bottom=20)

        def add_option(name):
            self.layout.add(OptionUI.create_group(self, name), fill=True,
                            margin_top=10, margin_bottom=10)

        # label = fsui.HeadingLabel(self, gettext("Keyboard Settings"))
        # self.layout.add(label, margin=10, margin_bottom=20)

        # add_option("automatic_input_grab")
        # add_option("initial_input_grab")
        add_option("keyboard_input_grab")
        add_option("swap_ctrl_keys")

        label = fsui.HeadingLabel(self, gettext("Key Mapping"))
        self.layout.add(label, margin_top=20, margin_bottom=20)

        add_option("keyboard_key_backslash")
        add_option("keyboard_key_equals")
        add_option("keyboard_key_insert")
        add_option("keyboard_key_less")



from fsbc.task import Task
import fsui as fsui
from fsui.extra.taskdialog import TaskDialog
from ...I18N import gettext
from fsui.extra.iconheader import NewIconHeader
from fsgs.context import fsgs
from fsgs.Database import Database
from fsgs.FileDatabase import FileDatabase
from fsgs.LockerDatabase import LockerDatabase
from fsgs.ogd.client import OGDClient


class MaintenanceSettingsPage(fsui.Panel):

    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        self.layout = fsui.VerticalLayout()
        # self.layout.set_padding(20, 20, 20, 20)

        self.icon_header = NewIconHeader(
            self, fsui.Icon("maintenance", "pkg:fs_uae_workspace"),
            gettext("Maintenance"),
            gettext("Miscellaneous functions to optimize {name}").format(
                name="FS-UAE Launcher"))
        self.layout.add(self.icon_header, fill=True, margin_bottom=20)

        # label = fsui.HeadingLabel(self, gettext("Defragment Databases"))
        # self.layout.add(label, fill=True, margin_top=20)

        # hori_layout = fsui.HorizontalLayout()
        # self.layout.add(hori_layout, fill=True, margin_top=10)
        # label = fsui.Label(self, gettext(
        #     "Defragmenting the databases will improve performance"))
        # hori_layout.add(label, expand=True)

        label = fsui.MultiLineLabel(self, gettext(
            "Defragmenting the databases will improve performance "
            "by ensuring that tables and indices are stored contiguously "
            "on disk. It will also reclaim some storage space."), 640)
        self.layout.add(label, fill=True, margin_top=20)

        button = fsui.Button(self, gettext("Defragment Databases"))
        button.activated.connect(self.on_defragment_button)
        # hori_layout.add(button, margin_left=20)
        self.layout.add(button, margin_top=20)

    def on_defragment_button(self):
        TaskDialog(self.get_window(), DefragmentDatabasesTask()).show()


class DefragmentDatabasesTask(Task):

    def __init__(self):
        Task.__init__(self, gettext("Defragment Databases"))

    def run(self):
        self.defragment(FileDatabase.get_instance(), "Files.sqlite")
        self.stop_check()
        self.defragment(LockerDatabase.instance(), "Locker.sqlite")
        self.stop_check()
        self.defragment(fsgs.get_game_database(),
                        OGDClient.get_server() + ".sqlite")
        self.stop_check()
        self.defragment(Database.get_instance(), "Database.sqlite")

    def defragment(self, database, name):
        self.set_progress(gettext("Defragmenting {name}").format(name=name))
        with database:
            cursor = database.cursor()
            cursor.execute("VACUUM")



import os
import time
import weakref
from uuid import uuid4
from fsgs.ogd.client import OGDClient
from urllib.request import urlopen
import threading
import traceback
import fsui as fsui
from fsgs.FSGSDirectories import FSGSDirectories
from ..Signal import Signal
from .Constants import Constants


SENTINEL = "3e31297d-2ae9-4014-9247-2495d40e5382"


class ImageLoader(object):

    instance = None

    @classmethod
    def get(cls):
        if not cls.instance:
            cls.instance = cls(SENTINEL)
        return cls.instance

    def __init__(self, sentinel):
        assert sentinel == SENTINEL
        self.stop_flag = False
        self.requests_lock = threading.Lock()
        self.requests = []
        threading.Thread(target=self.image_loader_thread,
                         name="ImageLoaderThread").start()

        Signal.add_listener("quit", self)

    def on_quit_signal(self):
        print("ImageLoader.on_quit_signal")
        self.stop_flag = True

    def image_loader_thread(self):
        try:
            self._image_loader_thread()
        except Exception:
            traceback.print_exc()

    def load_image(self, path="", sha1="", size=None, on_load=None, **kwargs):
        request = ImageLoadRequest()
        request.path = path
        request.sha1 = sha1
        request.image = None
        request.size = size
        request.on_load = on_load

        request.args = kwargs
        with self.requests_lock:
            self.requests.append(weakref.ref(request))
        return request

    def _image_loader_thread(self):
        while not self.stop_flag:
            # self.condition.wait()

            request = None
            with self.requests_lock:
                while len(self.requests) > 0:
                    request = self.requests.pop(0)()
                    if request is not None:
                        break
            if request:
                self.fill_request(request)
                request.notify()
            # FIXME: replace with condition
            time.sleep(0.01)

    def fill_request(self, request):
        try:
            self._fill_request(request)
        except Exception:
            traceback.print_exc()

    def get_cache_path_for_sha1(self, request, sha1):
        # print("get_cache_path_for_sha1", sha1)
        cover = request.args.get("is_cover", False)
        if cover:
            # size_arg = "?size={0}".format(256)
            # cache_ext = "_{0}".format(256)
            # size_arg = "?size={0}".format(128)
            # cache_ext = "_{0}".format(128)
            size_arg = "?w={0}&h={1}&t=lbcover".format(
                Constants.COVER_SIZE[0], Constants.COVER_SIZE[1])
            cache_ext = "{0}x{1}_lbcover.png".format(
                Constants.COVER_SIZE[0], Constants.COVER_SIZE[1])
        elif request.size:
            # size_arg = "?w={0}&h={1}".format(request.size[0],
            #         request.size[1])
            # cache_ext = "_{0}x{1}".format(request.size[0],
            #         request.size[1])
            size_arg = "?s=1x"
            cache_ext = "_1x.png"
        else:
            size_arg = ""
            cache_ext = ""

        cache_dir = os.path.join(
            FSGSDirectories.get_cache_dir(), "Images", sha1[:3])
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)
        cache_file = os.path.join(cache_dir, sha1 + cache_ext)
        if os.path.exists(cache_file):
            # an old bug made it possible for 0-byte files to exist, so
            # we check for that here..
            if os.path.getsize(cache_file) > 0:
                return cache_file

        server = OGDClient.get_server()
        url = "http://{}/image/{}{}".format(server, sha1, size_arg)
        print(url)
        r = urlopen(url)
        data = r.read()

        cache_file_partial = "{}.{}.partial".format(cache_file,
                                                    str(uuid4())[:8])
        with open(cache_file_partial, "wb") as f:
            f.write(data)
        os.rename(cache_file_partial, cache_file)
        return cache_file

    def _fill_request(self, request):
        if request.path is None:
            return
        cover = request.args.get("is_cover", False)
        
        if request.path.startswith("sha1:"):
            path = self.get_cache_path_for_sha1(request, request.path[5:])
        else:
            path = request.path

        if not path:
            return

        print("loading image from", request.path)
        image = fsui.Image(path)
        print(image.size, request.size)
        if request.size is not None:
            dest_size = request.size
        else:
            dest_size = image.size
        if image.size == dest_size:
            request.image = image
            return

        if cover:
            try:
                ratio = image.size[0] / image.size[1]
            except Exception:
                ratio = 1.0
            if 0.85 < ratio < 1.20:
                min_length = min(request.size)
                dest_size = (min_length, min_length)
            double_size = False
        else:
            double_size = True

        if double_size and image.size[0] < 400:
            image.resize(
                (image.size[0] * 2, image.size[1] * 2), fsui.Image.NEAREST)
        image.resize(dest_size)
        request.image = image


class ImageLoadRequest(object):

    def __init__(self):
        self.on_load = self._dummy_on_load_function
        self.size = None
        self.args = {}

    def notify(self):
        def on_load_function():
            self.on_load(self)
        fsui.call_after(on_load_function)

    def _dummy_on_load_function(self, obj):
        pass



import sys
import time
from fs_uae_launcher.ui.config.configscrollarea import ConfigScrollArea
import fstd.desktop
from fs_uae_launcher.ui.bottombar.GameInfoPanel import GameInfoPanel
from fs_uae_launcher.ui.bottombar.BottomPanel import BottomPanel
from fs_uae_launcher.ui.bottombar.ScreenshotsPanel import ScreenshotsPanel
from fs_uae_launcher.ui.bottombar.LaunchGroup import LaunchGroup
from fs_uae_workspace.shell import shell_open
import fsui as fsui
from fsbc.Application import Application
import fs_uae_launcher.ui
from ..Config import Config
from ..Signal import Signal
from fs_uae_launcher.Options import Option
from ..Settings import Settings
from ..I18N import gettext
from .AboutDialog import AboutDialog
from .Book import Book
from .CDPanel import CDPanel
from .ConfigurationsPanel import ConfigurationsPanel
from .Constants import Constants
from .config.additionalconfigpanel import AdditionalConfigPanel
from .config.expansionspanel import ExpansionsPanel
from .config.romrampanel import RomRamPanel
from .FloppiesPanel import FloppiesPanel
from .HardDrivesPanel import HardDrivesPanel
from .InputPanel import InputPanel
from .MainPanel import MainPanel
from .NetplayPanel import NetplayPanel
from .SetupDialog import SetupDialog
from .Skin import Skin
from .statusbar.StatusBar import StatusBar
from .WindowWithTabs import WindowWithTabs

USE_MAIN_MENU = 1


class MainWindow(WindowWithTabs):

    instance = None

    def __init__(self, fsgs, icon):
        self.fsgs = fsgs
        border = True
        maximize = None
        if fs_uae_launcher.ui.get_screen_size()[1] <= 768:
            if fstd.desktop.is_running_gnome_3():
                border = False
                if "--window-border" in sys.argv:
                    border = True
            maximize = True
        if "--no-window-border" in sys.argv:
            border = False
        title = "FS-UAE Launcher {0}".format(Application.instance().version)

        WindowWithTabs.__init__(self, None, title, border=border)
        if icon:
            self.set_icon_from_path(icon)

        self.tab_panels = []
        self.books = []
        self.menu_button = None
        self.main_menu_close_time = 0
        self.user_menu_close_time = 0
        self.user_button = None
        self.main_panel = None

        self.main_layout = fsui.VerticalLayout()
        self.set_content(self.main_layout)

        self.column_layout = fsui.HorizontalLayout()
        self.main_layout.add(self.column_layout, fill=True, expand=True)

        # left border
        self.create_column(
            0, min_width=Skin.get_window_padding_left(), content=False)
        # left content
        # if fs_uae_launcher.ui.get_screen_size()[0] > 1024:
        #     left_width = 518
        # else:
        #     left_width = 400
        # FIXME: adjust
        left_width = 462
        self.create_column(1, min_width=left_width)

        # right content
        # right_width = Constants.SCREEN_SIZE[0] * 2 + 21 + 10 + 10
        extra_screenshot_width = Constants.SCREEN_SIZE[0] + 21
        need_width = 1280

        right_width = 518

        if self.is_editor_enabled():
            # need_width += extra_screen_width
            pass
        else:
            if fs_uae_launcher.ui.get_screen_size()[0] >= need_width:
                # make room for one more screenshot
                right_width += extra_screenshot_width
                pass

        # right_width = 518

        # if Skin.EXTRA_GROUP_MARGIN:
        #     self.main_layout.add_spacer(Skin.EXTRA_GROUP_MARGIN)

        self.create_column(2, min_width=right_width, expand=True)

        # right border
        self.create_column(3, min_width=Skin.get_window_padding_right(),
                           content=False)

        # if self.is_editor_enabled():
        #     from ..editor.EditorGroup import EditorGroup
        #     editor = EditorGroup(self)
        #     self.main_layout.add(editor, fill=True, expand=True,
        #                          margin_right=20)

        # if fs_uae_launcher.ui.get_screen_size()[1] >= 768:
        #     right_margin = 0

        if fs_uae_launcher.ui.get_screen_size()[1] >= 720:
            self.bottom_layout = fsui.HorizontalLayout()
            self.main_layout.add(self.bottom_layout, fill=True)

            bottom_panel = BottomPanel(self)
            bottom_panel.set_min_width(10)
            bottom_panel.set_min_height(Skin.get_bottom_panel_height())
            self.bottom_layout.add(bottom_panel, fill=True)

            self.screenshots_panel = ScreenshotsPanel(self)
            self.bottom_layout.add(self.screenshots_panel, fill=True)

            self.game_info_panel = GameInfoPanel(self)
            self.game_info_panel.set_min_width(500)
            self.bottom_layout.add(self.game_info_panel, fill=True, expand=True)

            bottom_panel = BottomPanel(self)
            bottom_panel.set_min_width(10)
            bottom_panel.set_min_height(Skin.get_bottom_panel_height())
            self.bottom_layout.add(bottom_panel, fill=True)
        else:
            self.bottom_layout = None
            self.screenshots_panel = None

        # right_margin = -10 - Skin.EXTRA_GROUP_MARGIN
        #     else:
        #         bottom_panel = None
        #     # FIXME:
        #     if bottom_panel is None:
        #         layout.add_spacer(0, Skin.get_bottom_panel_height())
        #     else:
        #         bottom_panel.set_min_height(Skin.get_bottom_panel_height())
        #         layout.add(bottom_panel, fill=True, margin_right=right_margin)
        # elif column == 1:
        #     group = LaunchGroup(self)
        #     layout.add(group, fill=True, margin=10, margin_top=0)
        #     layout.add_spacer(0, 10)

        self.realize_tabs()
        # self.menu = self.create_menu()
        if fsui.System.macosx and fsui.toolkit == 'wx':
            # import wx
            # self.tools_menu = self.create_menu()
            # menu_bar = wx.MenuBar()
            # menu_bar.Append(self.tools_menu._menu, _("Tools"))
            # self.SetMenuBar(menu_bar)
            pass

        self.status_bar = StatusBar(self)
        self.layout.add(self.status_bar, fill=True)

        was_maximized = Settings.get("maximized") == "1"

        if Settings.get(Option.CONFIG_FEATURE) == "1":
            if fs_uae_launcher.ui.get_screen_size()[0] > 1300:
                if fs_uae_launcher.ui.get_screen_size()[1] > 1000:
                    self.layout.min_width = 1280
                    self.layout.min_height = 720

        self.set_size(self.layout.get_min_size())

        self.center_on_screen()
        if was_maximized or maximize:
            self.maximize()

        Signal.add_listener("scan_done", self)
        Signal.add_listener("setting", self)

        self.update_title()

    def on_destroy(self):
        print("MainWindow.destroy")
        Signal.remove_listener("scan_done", self)
        Signal.remove_listener("setting", self)

    def on_scan_done_signal(self):
        print("MainWindow.on_scan_done_signal")
        Config.update_kickstart()

    def update_title(self):
        auth = Settings.get(Option.DATABASE_AUTH)
        if auth:
            username = Settings.get(Option.DATABASE_USERNAME)
            login_info = username
        else:
            login_info = gettext("Not logged in")
        title = "FS-UAE Launcher {0} ({1})".format(
            Application.instance().version, login_info)
        self.set_title(title)

    def on_setting(self, key, value):
        if key == Option.DATABASE_AUTH:
            self.update_title()

    def on_resize(self):
        print("on_resize, size =", self.get_size(), self.is_maximized())
        if self.is_maximized():
            Settings.set("maximized", "1")
        else:
            Settings.set("maximized", "0")

        if self.screenshots_panel is not None:
            window_width = self.get_size()[0]
            available = (window_width - 10 - 10 -
                         self.game_info_panel.get_min_width())
            num_screenshots = int((available - 22 + 22) /
                                  (Constants.SCREEN_SIZE[0] + 22))
            print(num_screenshots)
            screenshots_panel_width = \
                (Constants.SCREEN_SIZE[0] + 22) * num_screenshots - 22 + 22
            self.screenshots_panel.set_min_width(screenshots_panel_width)

        fsui.Window.on_resize(self)

    def is_editor_enabled(self):
        return "--editor" in sys.argv

    def create_column(self, column, content=True, expand=False, min_width=0):
        layout = fsui.VerticalLayout()
        self.column_layout.add(layout, fill=True, expand=expand)
        if min_width:
            layout.add_spacer(min_width, 0)

        layout.add_spacer(0, 10 + Skin.EXTRA_GROUP_MARGIN)

        if content:
            book = Book(self)
            Skin.set_background_color(book)
            if column == 1:
                margin_right = Skin.EXTRA_GROUP_MARGIN
                # expand = False
                expand = True
            else:
                margin_right = 0
                expand = True
            hor_layout = fsui.HorizontalLayout()
            layout.add(
                hor_layout, fill=True, expand=expand, margin_right=margin_right)
            # hor_layout.add(
            #     book, fill=True, expand=expand, margin_right=margin_right)

            vert_layout = fsui.VerticalLayout()
            hor_layout.add(vert_layout, fill=True, expand=expand,
                           margin_right=margin_right)
            vert_layout.add(book, fill=True, expand=True)

#                        self.add_tab_panel(LaunchGroup, expand=False)

            if column == 2 and Settings.get(Option.CONFIG_FEATURE) == "1":
                if fs_uae_launcher.ui.get_screen_size()[0] >= 1280:
                    from fs_uae_launcher.ui.config.browser import ConfigBrowser
                    config_browser = ConfigBrowser(self)
                    config_browser.set_min_width(200)
                    hor_layout.add(config_browser, fill=True, expand=0.5,
                                   margin=10)

            if column == 2:
                # if fs_uae_launcher.ui.get_screen_size()[1] >= 1024:
                #     vert_layout.add_spacer(100)

                hor2_layout = fsui.HorizontalLayout()
                vert_layout.add(hor2_layout, fill=True, margin=10)
                launch_group = LaunchGroup(self)
                hor2_layout.add_spacer(0, expand=True)
                hor2_layout.add(launch_group)

            self.books.append(book)
            self.add_column_content(column)
        else:
            layout.add_spacer(0, expand=True)
            self.books.append(None)

        # if column == 1 and Settings.get(Option.CONFIG_FEATURE) == "1":
        #     from fs_uae_launcher.ui.config.browser import ConfigBrowser
        #     config_browser = ConfigBrowser(self)
        #     config_browser.set_min_height(200)
        #     layout.add(config_browser, fill=True, expand=True, margin=10)

        layout.add_spacer(0, 10 + Skin.EXTRA_GROUP_MARGIN)
        # if fs_uae_launcher.ui.get_screen_size()[1] >= 768:
        #     right_margin = 0
        #     if column == 0:
        #         bottom_panel = BottomPanel(self)
        #     elif column == 2:
        #         bottom_panel = GameInfoPanel(self)
        #     elif column == 1:
        #         bottom_panel = ScreenshotsPanel(self)
        #         right_margin = -10 - Skin.EXTRA_GROUP_MARGIN
        #     else:
        #         bottom_panel = None
        #     # FIXME:
        #     if bottom_panel is None:
        #         layout.add_spacer(0, Skin.get_bottom_panel_height())
        #     else:
        #         bottom_panel.set_min_height(Skin.get_bottom_panel_height())
        #         layout.add(bottom_panel, fill=True, margin_right=right_margin)
        # elif column == 1:
        #     group = LaunchGroup(self)
        #     layout.add(group, fill=True, margin=10, margin_top=0)
        #     layout.add_spacer(0, 10)

    def add_column_content(self, column):
        default_page_index = 0
        default_tab_index_offset = 0
        if column == 1:
            # if USE_MAIN_MENU:
            icon = fsui.Image("fs_uae_launcher:res/main_menu.png")
            self.menu_button = self.add_tab_button(
                None, icon, gettext("Main Menu"),
                menu_function=self.open_main_menu, left_padding=5)
            default_tab_index_offset = 1
            # self.add_tab_spacer(60)
            # else:
            #    self.add_tab_spacer(10)

            # page_index = 0
            self.add_page(column, ConfigurationsPanel, "",
                          gettext("Configurations"),
                          gettext("Configuration Browser"))

            # self.add_tab_spacer(194)

            # self.add_tab_spacer(10)
            # self.add_tab_spacer(10)

            # info_panel = self.add_tab_panel(InfoPanel, expand=False)
            # info_panel = \
            # self.add_tab_panel(InfoPanel, expand=True)
            # info_panel.set_min_width(316)
            # info_panel.set_min_width(326)
            # info_panel.set_min_width(272)

            self.add_tab_spacer(0, expand=True)

            # self.add_tab_spacer(10)
            # icon = fsui.Image("fs_uae_launcher:res/user_menu.png")
            # self.user_button = self.add_tab_button(
            #     None, icon, gettext("User Menu"),
            #     menu_function=self.open_user_menu,
            #     left_padding=5, right_padding=5)

            # self.add_tab_spacer(10)
            # if not USE_MAIN_MENU:
            #     icon = fsui.Image("fs_uae_launcher:res/tab_scan.png")
            #     self.add_tab_button(self.on_scan_button, icon,
            #                         gettext("Scan"),
            #                         gettext("Open Scan Dialog"))
            #     icon = fsui.Image("fs_uae_launcher:res/tab_settings.png")
            #     self.add_tab_button(self.on_settings_button, icon,
            #                         gettext("Settings"))

        elif column == 2:
            self.new_tab_group()

            self.add_tab_spacer(10)
            self.add_page(
                column, MainPanel,
                "crystal/32x32/apps/starthere",
                gettext("Config"), gettext("Main Configuration Options"))
            self.add_page(
                column, InputPanel,
                "tab_input_2",
                gettext("Input"), gettext("Input Options"))
            self.add_page(
                column, FloppiesPanel,
                "crystal/32x32/devices/3floppy_unmount_fs",
                gettext("Floppies"), gettext("Floppy Drives"))
            self.add_page(
                column, CDPanel,
                "crystal/32x32/devices/cdrom_unmount_fs",
                gettext("CD-ROMs"), gettext("CD-ROM Drives"))
            self.add_page(
                column, HardDrivesPanel,
                "crystal/32x32/apps/harddrive2",
                gettext("Hard Drives"))
            self.add_scroll_page(
                column, ExpansionsPanel,
                "crystal/32x32/apps/hardware_fs",
                gettext("Expansions"), gettext("Expansions"))
            self.add_scroll_page(
                column, RomRamPanel,
                "crystal/32x32/apps/kcmmemory",
                gettext("Hardware"), gettext("ROM and RAM"))
            self.add_scroll_page(
                column, AdditionalConfigPanel,
                "crystal/32x32/apps/kcontrol_fs",
                gettext("Additional Configuration"),
                gettext("Additional Configuration"))

            # self.add_tab_spacer(10)

            # if USE_MAIN_MENU:
            # self.add_tab_spacer(80)
            # else:
            #     icon = fsui.Image("fs_uae_launcher:res/tab_custom.png")
            #     self.add_tab_button(
            #         self.on_custom_button, icon, gettext("Custom"),
            #         gettext("Edit Custom Options"))
            #     self.add_tab_spacer(60)

            # self.add_tab_spacer(60)
            # self.add_tab_spacer(0, expand=True)
            # launch_group = LaunchGroup(self)
            # self.add_tab_panel(launch_group)
            # self.add_tab_panel(LaunchGroup)
            # self.add_tab_panel(LaunchGroup, expand=False)
            # self.add_tab_spacer(10)

            self.add_tab_spacer(0, expand=True)
            # FIXME: correct spacing so tabs are centered
            # self.add_tab_spacer(64)

            if Settings.get(Option.NETPLAY_FEATURE) == "1":
                # page_index += 1
                self.add_page(
                    column, NetplayPanel,
                    "crystal/32x32/apps/browser_fs",
                    gettext("Net Play"))
            self.add_tab_spacer(10)

        # column - 1 is the group id of the tab group
            self.select_tab(
                default_page_index + default_tab_index_offset, column - 1)
        self.books[column].set_page(default_page_index)

    def create_menu(self):
        menu = fsui.Menu()

        if Settings.get(Option.DATABASE_AUTH):
            menu.add_item(gettext("Update Game Database"),
                          self.on_update_game_database)
        menu.add_item(gettext("Update File Database"),
                      self.on_update_file_database)

        menu.add_separator()
        menu.add_item(gettext("ADF Creator") + "...", self.on_adf_creator)
        menu.add_item(gettext("HDF Creator") + "...", self.on_hdf_creator)

        menu.add_separator()
        menu.add_item(gettext("Import Kickstarts") + "...",
                      self.on_import_kickstarts)
        menu.add_item(gettext("Amiga Forever Import") + "...",
                      self.on_import_kickstarts)

        menu.add_separator()
        self.add_user_menu_content(menu)

        menu.add_separator()
        menu.add_preferences_item(
            gettext("Settings") + "...", self.on_settings_button)

        menu.add_separator()
        menu.add_item(gettext("About {name}").format(
            name="OAGD.net") + "...", self.on_what_is_this)
        menu.add_about_item(gettext("About {name}").format(
            name="FS-UAE Launcher") + "...", self.on_about)

        menu.add_separator()
        menu.add_about_item(gettext("Quit"), self.on_quit)
        return menu

    def _add_page(self, book, instance, icon_name, title, tooltip):
        book.add_page(instance)
        if icon_name:
            icon = fsui.Image("fs_uae_launcher:res/{0}.png".format(icon_name))
        else:
            icon = None

        def function():
            book.set_page(instance)

        if icon:
            self.add_tab(function, icon, title, tooltip)
        return instance

    def add_page(self, column, content_class, icon_name, title, tooltip=""):
        book = self.books[column]
        instance = content_class(book)
        if content_class == MainPanel:
            self.main_panel = instance
        return self._add_page(book, instance, icon_name, title, tooltip)

    def add_scroll_page(self, column, content_class, icon_name, title,
                        tooltip=""):
        book = self.books[column]
        instance = ConfigScrollArea(book)
        content_instance = content_class(instance)
        instance.set_widget(content_instance)
        if content_class == MainPanel:
            self.main_panel = instance
        return self._add_page(book, instance, icon_name, title, tooltip)

    def on_custom_button(self):
        from .config.ConfigDialog import ConfigDialog
        ConfigDialog.run(self.get_window())

    def open_main_menu(self):
        if fsui.System.windows:
            if time.time() - getattr(self, "main_menu_close_time", 0) < 0.2:
                return
        self.menu = self.create_menu()
        if Skin.use_unified_toolbar():
            self.popup_menu(self.menu, (0, -2))
        else:
            self.menu_button.popup_menu(
                self.menu, (0, self.menu_button.size[1] - 2))
        if fsui.System.windows:
            self.main_menu_close_time = time.time()

    # def open_user_menu(self):
    #     if fsui.System.windows:
    #         if time.time() - getattr(self, "user_menu_close_time", 0) < 0.2:
    #             return
    #     user_menu = self.create_user_menu()
    #     self.user_button.popup_menu(
    #         user_menu, (0, self.user_button.get_size()[1] - 2))
    #     if fsui.System.windows:
    #         self.user_menu_close_time = time.time()

    def add_user_menu_content(self, menu):
        if Settings.get(Option.DATABASE_AUTH):
            # menu.add_item(_("Log In / Register"), self.on_log_in)
            # menu.add_item(gettext("Update Game Database"),
            #               self.on_game_database_refresh)
            menu.add_item(gettext("Upload Files to OAGD.net Locker"),
                          self.on_upload_locker_files)
            # menu.add_separator()
            menu.add_item(gettext("Log Out"), self.on_log_out)
        else:
            menu.add_item(gettext("Log In / Register"), self.on_log_in)

    def create_user_menu(self):
        menu = fsui.Menu()
        self.add_user_menu_content(menu)
        return menu

    def on_log_in(self):
        print("on_log_in")
        shell_open("Workspace:Prefs/User/Login", parent=self)

    def on_log_out(self):
        print("on_log_out")
        shell_open("Workspace:Prefs/User/Logout", parent=self)

    def on_what_is_this(self):
        print("on_what_is_this")
        fstd.desktop.open_url_in_browser("http://oagd.net/about")

    def on_scan_button(self):
        from .ScanDialog import ScanDialog
        ScanDialog(self.get_window()).show()

    def on_settings_button(self):
        from .settings.SettingsDialog import SettingsDialog
        SettingsDialog.open(self)

    def on_about(self):
        dialog = AboutDialog(self.get_window())
        dialog.show()

    def on_quit(self):
        self.close()

    def on_import_kickstarts(self):
        SetupDialog(self.get_window()).show()

    def on_update_file_database(self):
        self.on_scan_button()

    def on_update_game_database(self):
        shell_open("Workspace:Tools/Refresh", parent=self)

    def on_upload_locker_files(self):
        print("on_upload_locker_files")
        shell_open("Workspace:Tools/LockerUploader", parent=self)

    def on_adf_creator(self):
        print("on_adf_creator")
        shell_open("Workspace:Tools/ADFCreator", parent=self)

    def on_hdf_creator(self):
        print("on_hdf_creator")
        shell_open("Workspace:Tools/HDFCreator", parent=self)



import uuid
import traceback
import random
from urllib.request import urlopen, URLError
from urllib.parse import parse_qs
from fsgs.context import fsgs
from .IRC import IRC
from .Channel import Channel
from .IRCColor import IRCColor
from .IRCBroadcaster import IRCBroadcaster
from .ConnectionTester import ConnectionTester
from fsgs.amiga.Amiga import Amiga
from ..Config import Config
from ..server.Server import Server


class Netplay:

    host_ports = []

    enabled = False
    game_channel = ""
    connection_tester = None
    start_sequence = ""
    players = {}

    @classmethod
    def new_host_port(cls):
        if cls.host_ports is not None:
            cls.host_ports = list(range(25102, 25500))
            random.shuffle(cls.host_ports)
        return cls.host_ports.pop()

    @classmethod
    def player(cls, nick):
        try:
            return cls.players[nick]
        except Exception:
            from .Player import Player
            cls.players[nick] = Player(nick)
            print("players are now", cls.players)
            return cls.players[nick]

    @classmethod
    def remove_player(cls, nick):
        try:
            del cls.players[nick]
        except Exception:
            print(
                "warning, tried to remove player (but wasn't in list):", nick)
        else:
            print("players are now", cls.players)

    @classmethod
    def start_netplay_game(cls):
        if Netplay.set_ready():
            if Netplay.is_op():
                cls.game_info("you are marked as ready, initiating game start")
                Netplay.initiate_start_sequence()
            else:
                cls.game_info(
                    "you are marked as ready (only ops can start game)")
        else:
            cls.game_info("you are not ready")

    @classmethod
    def is_op(cls):
        return IRC.channel(cls.game_channel).is_op(IRC.my_nick)

    @classmethod
    def initiate_start_sequence(cls):
        print("initiate start sequence")
        if not cls.game_channel:
            cls.game_info("cannot start - no game channel")
        cls.start_sequence = str(uuid.uuid4())
        message = "__prestart {0} {1}".format(
            cls.start_sequence, cls.get_config_hash())
        IRC.channel(cls.game_channel).privmsg(message)
        # in case one wants to try a single-player game for testing
        cls.on_ackstart()

    @classmethod
    def on_ackstart(cls):
        config_hash = cls.get_config_hash()
        start_sequence = cls.start_sequence
        player_count = 0
        for player in cls.players.values():
            player_count += 1
            if player.nick == IRC.my_nick:
                # implicitly ok!
                continue
            if player.get("config_hash") != config_hash:
                return
            if player.get("start_sequence") != start_sequence:
                return
        want_players = int(Config.get("__netplay_players") or "0")
        if player_count != want_players:
            cls.game_warning("cannot start game, wanted {0} players, "
                             "has {1}".format(want_players, player_count))
            return
        cls.game_info("sending game start command to all clients!")
        message = "__start {0} {1}".format(
            cls.start_sequence, cls.get_config_hash())
        IRC.channel(cls.game_channel).privmsg(message)
        cls.do_start_game()

    @classmethod
    def do_start_game(cls):
        cls.game_info("starting game!")
        print("do start game")

        # we now reset __netplay_addresses so that the ConnectionTester
        # does not connect to the game in progress
        Config.set("__netplay_addresses", "")
        from ..FSUAELauncher import FSUAELauncher
        FSUAELauncher.start_local_game()

        # game done, resetting config as the same game server cannot be
        # used again
        cls.reset_netplay_config()

    @classmethod
    def get_config_hash(cls):
        return Config.checksum()

    @classmethod
    def set_ready(cls):
        if not cls.enabled:
            return False
        if not cls.game_channel:
            IRC.warning("cannot set ready - not in a game")
            return False
        if not Config.get("__netplay_addresses"):
            cls.game_warning("cannot set ready - no server configured")
            cls.game_info("an operator needs to use /startgame or "
                          "/hostgame (or /customgame) first")
            return False
        if not Config.get("__netplay_host"):
            cls.game_warning("cannot set ready - no connection to game "
                             "server")
            return False
        Config.set("__netplay_ready", "1")
        # check in case an component has force __netplay_ready back to 0
        return Config.get("__netplay_ready") == "1"

    @classmethod
    def connect(cls):
        cls.enabled = True
        IRC.connect()
        if cls.connection_tester is None:
            cls.connection_tester = ConnectionTester()

    @classmethod
    def is_connected(cls):
        return cls.enabled

    @classmethod
    def on_config(cls, key, value):
        if key == "__netplay_host_last_error" and value:
            cls.game_warning("problem connecting to game server: " + value)
        elif key == "__netplay_host":
            if value:
                cls.game_info("successfully connected to net play host: " +
                              value)
        if not IRC.running:
            return
        if not cls.game_channel:
            return
        if key in Config.sync_keys_set:
            message = "__config {0} {1}".format(key, value)
            IRC.channel(cls.game_channel).privmsg(message)

    @classmethod
    def game_message(cls, message, color=None):
        if not cls.game_channel:
            IRC.warning("game message received (but not "
                        "currently in a game channel): " + message)
            return
        IRC.channel(cls.game_channel).message(message, color)
        if cls.game_channel != IRC.active_channel:
            # also paste message to active console
            Channel.active().message(message, color)

    @classmethod
    def game_warning(cls, message):
        cls.game_message(message, IRCColor.WARNING)

    @classmethod
    def game_info(cls, message):
        cls.game_message(message, IRCColor.INFO)

    @classmethod
    def is_game_channel(cls, name):
        if not IRC.is_channel(name):
            return False
        if name.startswith("&"):
            return True
        if name.endswith("-game"):
            return True
        return False

    @classmethod
    def on_irc(cls, key, args):
        if key == "joined":
            if cls.is_game_channel(args["channel"]):
                if cls.game_channel:
                    IRC.part(cls.game_channel)
                cls.game_channel = ""
                cls.reset_netplay_config()
                cls.on_join_game_channel(args["channel"])
        elif key == "parted":
            if cls.game_channel == args["channel"]:
                cls.game_channel = ""
                cls.reset_netplay_config()
        elif key == "privmsg":
            if args["message"].startswith("__"):
                if args["channel"] == cls.game_channel:
                    cls.handle_game_instruction(args["nick"], args["message"])
        elif key == "join":
            if args["channel"] == cls.game_channel:
                # initialize new player
                cls.player(args["nick"])
                if args["nick"] != IRC.my_nick:
                    if cls.is_op():
                        # operator - send config to new player(s)
                        IRC.channel(
                            cls.game_channel).action(
                                "sends config on arrival of new player")
                        cls.send_config()
        elif key == "part":
            if args["channel"] == cls.game_channel:
                Netplay.remove_player(args["nick"])

    @classmethod
    def on_join_game_channel(cls, channel):
        # first we erase config which is not part of the default configuration
        # key set, to avoid some problems
        keys = fsgs.config.values.keys()
        for key in keys:
            if key not in Config.default_config:
                Config.set(key, "")
        cls.game_channel = channel
        # reset list of players
        cls.players = {}

    @classmethod
    def reset_config(cls):
        Config.load_default_config()

    @classmethod
    def reset_netplay_config(cls):
        Config.set_multiple([
            ("__netplay_id", ""),
            ("__netplay_password", ""),
            ("__netplay_players", ""),
            ("__netplay_port", ""),
            ("__netplay_addresses", ""),
            ("__netplay_host", "")])

    @classmethod
    def handle_command(cls, command):
        if not command.startswith("/"):
            return False
        args = command[1:].split(" ")
        print(args)
        function_name = "command_" + args[0]
        try:
            function = getattr(cls, function_name)
        except AttributeError:
            return False
        else:
            try:
                function(args[1:])
            except Exception:
                cls.game_warning(traceback.format_exc())
            return True

    @classmethod
    def command_ready(cls, args):
        if len(args) != 0:
            IRC.warning("syntax: /ready")
            return
        cls.set_ready()
        if Config.get("__netplay_ready") == "1":
            cls.game_info("status set to ready")

    @classmethod
    def command_notready(cls, args):
        if len(args) != 0:
            IRC.warning("syntax: /notready")
            return
        Config.set("__netplay_ready", "0")
        cls.game_info("status set to not ready")

    @classmethod
    def command_check(cls, args):
        channel = Channel.active()
        if not cls.require_game_channel(channel):
            return
        if len(args) != 1:
            IRC.warning("syntax: /check <key>")
            return
        if not channel.is_op():
            IRC.warning("check: you need to be an operator")
            return
        for arg in args:
            cls.print_check_request(IRC.my_nick, arg, Config.get(arg))
            message = "__check {0} {1}".format(arg, Config.get(arg))
            IRC.channel(cls.game_channel).privmsg(message)

    @classmethod
    def command_verify(cls, args):
        channel = Channel.active()
        if not cls.require_game_channel(channel):
            return
        if len(args) != 0:
            IRC.warning("syntax: /verify")
            return
        if not channel.is_op():
            IRC.warning("verify: you need to be an operator")
            return
        # for arg in args:
        #     cls.game_info("requesting verification of {0} = {1}".format(arg,
        #             Config.get(arg)))
        cls.print_verify_request(IRC.my_nick)
        IRC.channel(cls.game_channel).privmsg("__beginverify")
        for key in Config.checksum_keys:
            value = Config.get(key)
            message = "__verify {0} {1}".format(key, value)
            IRC.channel(cls.game_channel).privmsg(message)
        IRC.channel(cls.game_channel).privmsg("__endverify")

    @classmethod
    def command_sendconfig(cls, args):
        channel = Channel.active()
        if not cls.require_game_channel(channel):
            return
        if len(args) != 0:
            IRC.warning("syntax: /sendconfig")
            return
        cls.game_info("sending config")
        cls.send_config()

    @classmethod
    def send_config(cls):
        if not cls.game_channel:
            return
        channel = IRC.channel(cls.game_channel)
        # sending keys in preferred key order
        for key in Config.sync_keys_list:
            value = Config.get(key)
            message = "__config {0} {1}".format(key, value)
            channel.privmsg(message)

    @classmethod
    def require_game_channel(cls, channel):
        if not IRC.running:
            return False
        result = cls.game_channel and channel.name == cls.game_channel
        if not result:
            channel.warning("not in active game channel")
        return result

    @classmethod
    def command_resetconfig(cls, args):
        if len(args) != 0:
            IRC.warning("syntax: /resetconfig")
            return
        channel = Channel.active()
        if not cls.require_game_channel(channel):
            return

        channel.info("config reset by request")
        # send resetconfig command to other players first
        if channel.is_op():
            channel.privmsg("__resetconfig")
        # then reset own config, so config changes are applied after
        # other players have reset their configs
        Netplay.reset_config()

    @classmethod
    def parse_server_args(cls, args, default_port):
        if len(args) == 0:
            return None
        host = args.pop(0)
        if ":" in host:
            host, port = host.split(":", 1)
            port = int(port)
        else:
            port = default_port
        if len(args) > 0:
            players = args.pop(0)
            players = int(players)
        else:
            players = len(cls.players)
        if len(args) > 0:
            password = args.pop(0)
        else:
            password = ""
        if len(args) > 0:
            return None
        return host, port, players, password

    @classmethod
    def command_startserver(cls, args):
        return cls.command_startgame(args)

    @classmethod
    def command_startgame(cls, args):
        channel = Channel.active()
        if not cls.require_game_channel(channel):
            return
        result = cls.parse_server_args(args, 25101)
        if not result:
            IRC.warning("usage: /startgame <host>[:<port>] "
                        "[<players>] [<password>]")
            return
        host, port, players, password = result

        if not channel.is_op():
            IRC.warning("startgame: you need to be an operator")
            return

        url = "http://{0}:{1}/game/create?players={2}&password={3}".format(
            host, port, players, password)
        try:
            f = urlopen(url)
            result = f.read()
        except URLError:
            channel.warning(
                "Problem starting game server: {0}".format(
                    traceback.format_exc()))
        else:
            result = result.decode("UTF-8")
            print(result)
            result_dict = parse_qs(result)
            print(result_dict)
            game_id = result_dict["id"][0]
            game_password = result_dict["password"][0]
            game_port = result_dict["port"][0]
            game_addresses = result_dict["addresses"][0]
            Config.set_multiple([
                ("__netplay_game", game_id),
                ("__netplay_password", game_password),
                ("__netplay_players", str(players)),
                ("__netplay_port", game_port),
                ("__netplay_host", ""),
                ("__netplay_addresses", game_addresses)])
            channel.info("started game id: {0} password: {1} "
                         "server: {2} port: {3}".format(
                             game_id, game_password, game_addresses,
                             game_port))

    @classmethod
    def command_hostgame(cls, args):
        channel = Channel.active()
        if not cls.require_game_channel(channel):
            return
        result = cls.parse_server_args(args, cls.new_host_port())
        if not result:
            IRC.warning("usage: /hostgame <address>[,<address2>][:<port>] "
                        "[<players>] [<password>]")
            return
        addresses, port, players, password = result

        if not channel.is_op():
            IRC.warning("hostgame: you need to be an operator")
            return

        server = Server(port, players, password)
        server.start()

        from ..server.ServerWindow import ServerWindow
        window = ServerWindow(None, server)
        window.show()

        game_id = str(uuid.uuid4())
        Config.set_multiple([
            ("__netplay_game", game_id),
            ("__netplay_password", password),
            ("__netplay_players", str(players)),
            ("__netplay_port", str(port)),
            ("__netplay_host", ""),
            ("__netplay_addresses", addresses)])
        channel.info("started game id: {0} password: {1} "
                     "server: {2} port: {3}".format(
                         game_id, password, addresses, port))

    @classmethod
    def command_setserver(cls, args):
        return cls.command_customgame(args)

    @classmethod
    def command_customgame(cls, args):
        channel = Channel.active()
        if not cls.require_game_channel(channel):
            return
        result = cls.parse_server_args(args, 25100)
        if not result:
            IRC.warning("usage: /customgame <address>[,<address2>][:<port>] "
                        "[<players>] [<password>]")
            return
        addresses, port, players, password = result

        if not channel.is_op():
            IRC.warning("customgame: you need to be an operator")
            return

        game_id = str(uuid.uuid4())
        Config.set_multiple([
            ("__netplay_game", game_id),
            ("__netplay_password", password),
            ("__netplay_players", str(players)),
            ("__netplay_port", str(port)),
            ("__netplay_host", ""),
            ("__netplay_addresses", addresses)])
        channel.info("started game id: {0} password: {1} "
                     "server: {2} port: {3}".format(
                         game_id, password, addresses, port))

    @classmethod
    def command_set(cls, args):
        channel = Channel.active()
        if not cls.require_game_channel(channel):
            return
        if len(args) < 1:
            IRC.warning("usage: /set <key> [<value>]")
            return
        if not channel.is_op():
            IRC.warning("hostgame: you need to be an operator")
            return

        key = args[0]
        value = " ".join(args[1:])
        print("config, setting", key, "to", value)
        channel.action("set option {0} to \"{1}\"".format(key, value))
        Config.set(key, value)

    @classmethod
    def handle_game_instruction(cls, nick, message):
        channel = IRC.channel(cls.game_channel)

        # if Netplay.game_channel != self.name:
        #     self.warning(self, "ignored command in "
        #             "non-active game: {0}".format(command), IRCColor.WARNING)
        #     return
        words = message.split(" ")
        command = words[0]
        arg = " ".join(words[1:])
        if command == "__config":
            if channel.is_op(nick):
                args = arg.split(" ", 1)
                key = args[0]
                value = ""
                if len(args) > 1:
                    value = args[1]
                cls.set_config(key.strip(), value.strip())
        elif command == "__prestart":
            if channel.is_op(nick):
                seq, config_hash = arg.split(" ")
                my_config_hash = cls.get_config_hash()
                if my_config_hash != config_hash:
                    channel.privmsg("not the same config hash")
                    return
                if Config.get("__netplay_ready") != "1":
                    channel.privmsg("(op tried to start game, "
                                    "but I am not ready)")
                    return
                channel.privmsg("__ackstart {0} {1}".format(
                    seq, my_config_hash))
        elif command == "__ackstart":
            start_sequence, config_hash = arg.split(" ")
            cls.player(nick).set("start_sequence", start_sequence)
            cls.player(nick).set("config_hash", config_hash)
            cls.on_ackstart()
            # FIXME: check start sequence?
        elif command == "__start":
            if channel.is_op(nick):
                start_sequence, config_hash = arg.split(" ")
                my_config_hash = Netplay.get_config_hash()
                if my_config_hash != config_hash:
                    channel.action("could not start game "
                                   "(mismatching config hash)")
                    return
                Netplay.do_start_game()
        elif command == "__resetconfig":
            if channel.is_op(nick):
                channel.info("config reset by op")
                Netplay.reset_config()
        elif command == "__check":
            if channel.is_op(nick):
                args = arg.split(" ")
                key = args[0]
                value = " ".join(args[1:])
                my_value = Config.get(key)
                cls.print_check_request(nick, key, value)
                cls.print_check_response(IRC.my_nick, key, my_value, value)
                channel.privmsg("__ackcheck {0} {1}".format(key, my_value))
        elif command == "__beginverify":
            if channel.is_op(nick):
                cls.print_verify_request(nick)
        elif command == "__verify":
            if channel.is_op(nick):
                args = arg.split(" ")
                key = args[0]
                value = " ".join(args[1:])
                if Config.get(key) != value:
                    cls.print_verify_response(
                        IRC.my_nick, key, Config.get(key))
                    channel.privmsg("__ackverify {0} {1}".format(
                        key, Config.get(key)))
        elif command == "__endverify":
            pass
        elif command == "__ackcheck":
            args = arg.split(" ")
            key = args[0]
            value = " ".join(args[1:])
            my_value = Config.get(key)
            cls.print_check_response(nick, key, value, my_value)
        elif command == "__ackverify":
            args = arg.split(" ", 1)
            cls.print_verify_response(nick, args[0], " ".join(args[1:]))
        else:
            channel.warning(
                "unknown command received: {0}".format(command))

    @classmethod
    def print_check_request(cls, nick, key, value):
        IRC.channel(
            cls.game_channel).info(
                "* {0} requested check of {1} ({2})".format(
                    nick, key, value))

    @classmethod
    def print_check_response(cls, nick, key, value, check_value):
        if value == check_value:
            color = IRCColor.INFO
        else:
            color = IRCColor.WARNING
        IRC.channel(
            cls.game_channel).message(
                "{0} = {1} ({2})".format(key, value, nick), color)

    @classmethod
    def print_verify_request(cls, nick):
        IRC.channel(
            cls.game_channel).info(
                "* {0} requested automatic config verification".format(nick))

    @classmethod
    def print_verify_response(cls, nick, key, value):
        IRC.channel(
            cls.game_channel).warning(
                "{0} = {1} ({2})".format(key, value, nick))

    @classmethod
    def set_config(cls, key, value):
        # this config was received from a game channel operator
        print("received config", key, value)
        # channel =
        IRC.channel(cls.game_channel)
        if key not in Config.sync_keys_set:
            print("not processing this key")
            return
        elif key in ["x_kickstart_file_sha1", "x_kickstart_ext_file_sha1"]:
            cls.set_kickstart_config(key, value)
        elif key in cls.file_config:
            cls.set_file_config(key, value)
        else:
            Config.set(key, value)

    @classmethod
    def set_kickstart_config(cls, key, value):
        channel = IRC.channel(cls.game_channel)
        if key == "x_kickstart_file_sha1":
            if value == Amiga.INTERNAL_ROM_SHA1:
                    Config.set_multiple([
                        ("kickstart_file", "internal"),
                        ("x_kickstart_file", "internal"),
                        (key, value)])
            else:
                path = fsgs.file.find_by_sha1(sha1=value)
                if path:
                    Config.set_multiple([
                        ("kickstart_file", path),
                        ("x_kickstart_file", path),
                        (key, value)])
                else:
                    channel.action("could not find kickstart for {0}".format(
                                   repr(value)))
        elif key == "x_kickstart_ext_file_sha1":
            if not value:
                Config.set_multiple([
                    ("kickstart_ext_file", ""),
                    ("x_kickstart_ext_file", ""),
                    (key, "")])
                return
            path = fsgs.file.find_by_sha1(sha1=value)
            if path:
                Config.set_multiple([
                    ("kickstart_ext_file", path),
                    ("x_kickstart_ext_file", path),
                    (key, value)])
            else:
                channel.action("could not find (ext) kickstart "
                               "for {0}".format(repr(value)))

    @classmethod
    def set_file_config(cls, key, value):
        channel = IRC.channel(cls.game_channel)
        set_key = cls.file_config[key]
        if not value:
            Config.set_multiple([(set_key, ""), (key, "")])
            return
        if value.startswith("http://") or value.startswith("https://"):
            path = value
        else:
            # path = Database.get_instance().find_file(sha1=value)
            path = fsgs.file.find_by_sha1(sha1=value)
        # if not path:
        #     path = find_downloadable_file(value)
        if path:
            Config.set_multiple([(set_key, path), (key, value)])
        else:
            Config.set_multiple([(set_key, ""), (key, "")])
            channel.action("could not find {1} for "
                           "for {0}".format(value, set_key))

    file_config = {
    }
    for i in range(Amiga.MAX_FLOPPY_DRIVES):
        file_config["x_floppy_drive_{0}_sha1".format(i)] = \
            "floppy_drive_{0}".format(i)
    for i in range(Amiga.MAX_FLOPPY_IMAGES):
        file_config["x_floppy_image_{0}_sha1".format(i)] = \
            "floppy_image_{0}".format(i)
    for i in range(Amiga.MAX_CDROM_DRIVES):
        file_config["x_cdrom_drive_{0}_sha1".format(i)] = \
            "cdrom_drive_{0}".format(i)
    for i in range(Amiga.MAX_CDROM_IMAGES):
        file_config["x_cdrom_image_{0}_sha1".format(i)] = \
            "cdrom_image_{0}".format(i)
    for i in range(Amiga.MAX_HARD_DRIVES):
        file_config["x_hard_drive_{0}_sha1".format(i)] = \
            "hard_drive_{0}".format(i)


Config.add_listener(Netplay)
IRCBroadcaster.add_listener(Netplay)



from .mess import MESSRunner


class MegaDriveRunner(MESSRunner):

    CONTROLLER = {
        "type": "gamepad",
        "description": "Gamepad",
        "mapping_name": "megadrive",
    }

    PORTS = [
        {
            "description": "1st Controller",
            "types": [CONTROLLER]
        }, {
            "description": "2nd Controller",
            "types": [CONTROLLER]
        },
    ]

    # self.inputs.append(self.create_input(
    #         name='Controller {0}'.format(i + 1),
    #         type='megadrive',
    #         description='Gamepad'))

    def mess_configure(self):
        self.mess_configure_cartridge()

    def mess_input_mapping(self, _):
        return {
            "START": "P#_START",
            "MODE": "P#_SELECT",
            "UP": "P#_JOYSTICK_UP",
            "DOWN": "P#_JOYSTICK_DOWN",
            "LEFT": "P#_JOYSTICK_LEFT",
            "RIGHT": "P#_JOYSTICK_RIGHT",
            "A": "P#_BUTTON1",
            "B": "P#_BUTTON2",
            "C": "P#_BUTTON3",
            "X": "P#_BUTTON4",
            "Y": "P#_BUTTON5",
            "Z": "P#_BUTTON6",
        }

    def mess_romset(self):
        if self.is_pal():
            return "megadriv", {}
        else:
            return "genesis", {}

    # def mess_get_firmware_name(self):
    #     return None



import fsui
from fs_uae_launcher.I18N import gettext
from fs_uae_launcher.Options import Option
from fs_uae_launcher.ui.config.configpanel import ConfigPanel
from .ConfigDialog import ConfigDialog


class AdditionalConfigPanel(ConfigPanel):

    def __init__(self, parent):
        ConfigPanel.__init__(self, parent)

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)

        heading_label = fsui.HeadingLabel(
            self, gettext("Additional Configuration"))
        hori_layout.add(heading_label, margin=10)
        hori_layout.add_spacer(0, expand=True)

        self.custom_button = fsui.Button(
            self, gettext("Custom Configuration") + "...")
        self.custom_button.activated.connect(self.on_custom_button)
        hori_layout.add(self.custom_button, margin_right=10)

        self.layout.add_spacer(0)

        self.add_option(Option.CPU)
        self.add_option(Option.FLOPPY_DRIVE_SPEED)
        self.add_option(Option.FLOPPY_DRIVE_COUNT)
        self.add_option(Option.CDROM_DRIVE_COUNT)
        # self.add_option(Option.DONGLE_TYPE)

        self.layout.add_spacer(10)
        heading_label = fsui.HeadingLabel(self, gettext("Deprecated Options"))
        self.layout.add(heading_label, margin=10)
        self.layout.add_spacer(0)

        self.add_option(Option.UAEGFX_CARD)

    def on_custom_button(self):
        ConfigDialog.run(self.get_window())



import fsui as fsui
from .ExpansionsGroup import ExpansionsGroup
from ..Skin import Skin
from fs_uae_launcher.Options import Option
from .configoptionui import ConfigOptionUI


class ConfigPanel(fsui.Panel):

    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        Skin.set_background_color(self)

        # self.panel = fsui.Panel(self)
        # self.set_widget(self.panel)

        # self.expansions_group = ExpansionsGroup(self.panel)

        self.layout = fsui.VerticalLayout()
        # self.panel.layout.add(self.expansions_group, fill=True)

    def add_option(self, name):
        self.layout.add(ConfigOptionUI.create_group(self, name),
                        fill=True, margin=10)



from fsgs.platform import PlatformHandler
from fsgs.mess.atari_5200 import Atari5200Runner
from .loader import SimpleLoader


class Atari5200PlatformHandler(PlatformHandler):

    PLATFORM_NAME = "Atari 5200"

    def __init__(self):
        PlatformHandler.__init__(self)

    def get_loader(self, fsgs):
        return Atari5200Loader(fsgs)

    def get_runner(self, fsgs):
        return Atari5200Runner(fsgs)


class Atari5200Loader(SimpleLoader):
    pass



import fsui as fsui
from .config.WHDLoadGroup import WHDLoadGroup
from .config.HardDriveGroup import HardDriveGroup
from .Skin import Skin


class HardDrivesPanel(fsui.Panel):

    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        Skin.set_background_color(self)

        self.hard_drive_groups = []
        for i in range(4):
            self.hard_drive_groups.append(HardDriveGroup(self, i))
        self.whdload_group = WHDLoadGroup(self)

        self.layout = fsui.VerticalLayout()
        for i in range(4):
            self.layout.add(self.hard_drive_groups[i], fill=True, margin=10)
        self.layout.add_spacer(Skin.EXTRA_GROUP_MARGIN)
        self.layout.add(self.whdload_group, fill=True)



from fsui.qt import QSize, QColor, QLabel, QHBoxLayout
from fsui.qt import QMainWindow, QWidget, QVBoxLayout
from fs_uae_workspace.vfs import VFSDesktopItem

_root_window = None
_desktop_window = None


class DesktopWindow(QMainWindow):

    def __init__(self):
        QMainWindow.__init__(self)
        # .TopToBottom, self)

        self.root = QWidget(self)
        self.setCentralWidget(self.root)

        # self.root.setAutoFillBackground(True)
        # p = self.root.palette()
        # p.setColor(self.root.backgroundRole(), QColor(0xaa, 0xaa, 0xaa))
        # self.root.setPalette(p)

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        # self.root.setSizePolicy(QSizePolicy(2, 2), QSizePolicy(2, 2))
        # layout.addWidget(self.root, 1)
        # self.setLayout(layout)

        self.setMinimumSize(QSize(1600, 900))
        self.show()
        # self.showMaximized()

        self.top = QWidget()
        self.top.setAutoFillBackground(True)
        p = self.top.palette()
        p.setColor(self.top.backgroundRole(), QColor(0xff, 0xff, 0xff))
        self.top.setPalette(p)
        # self.top.setStyleSheet("background-color: #ffffff;")
        self.top.setMinimumHeight(26)
        layout.addWidget(self.top)

        workspace_label = QLabel("Workspace")
        top_layout = QHBoxLayout()
        top_layout.setContentsMargins(10, 0, 0, 0)
        top_layout.addWidget(workspace_label)
        self.top.setLayout(top_layout)

        # self.top_padding = QWidget()
        # self.top_padding.setAutoFillBackground(True)
        # p = self.top_padding.palette()
        # p.setColor(self.top_padding.backgroundRole(),
        #            QColor(0xaa, 0xaa, 0xaa))
        # self.top_padding.setPalette(p)
        # self.top_padding.setMinimumHeight(26)
        # layout.addWidget(self.top_padding)

        self.root_item = VFSDesktopItem()
        from fs_uae_workspace.vfsui import VFSIconView
        self.icon_view = VFSIconView(None, self.root_item)
        self.icon_view.show()
        self.icon_view.resize(20, 20)

        layout.addWidget(self.icon_view)

        self.root.setLayout(layout)

    def get_root_window(self):
        return self.root


def get_desktop_window():
    global _desktop_window
    if not _desktop_window:
        _desktop_window = DesktopWindow()
    return _desktop_window


def get_root_window():
    return get_desktop_window().get_root_window()



from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import os
from fsbc.system import windows, macosx


class GameCenter(object):
    game_dirs = []
    data_dir = None
    save_dir = None

    @classmethod
    def init(cls):
        print("GameCenter.init")
        # for arg in sys.argv:
        #     if arg.startswith("--game-dir="):
        #         cls.game_dirs.append(arg[11:])
        # k = 0
        # path = fs.get_lib_dir()
        # p = os.path.dirname(path)
        # while p != path:
        #     if fs.normalize_path(p) == \
        #             fs.normalize_path(get_home_dir()):
        #         break
        #     if os.path.exists(os.path.join(p, "Games")):
        #         cls.game_dirs.append(os.path.join(p, "Games"))
        #         print("found 'local' game dir ", repr(cls.game_dirs[-1]))
        #         break
        #     path = p
        #     p = os.path.dirname(path)
        # while p != path:
        #     if os.path.exists(os.path.join(p, "Data", "FSGameCenter")):
        #         cls.data_dir = os.path.join(p, "Data", "FSGameCenter")
        #         print("found 'local' data dir ", repr(cls.data_dir))
        #         break
        #     path = p
        #     p = os.path.dirname(path)
        # while p != path:
        #     if os.path.exists(os.path.join(p, "Save")):
        #         cls.save_dir = os.path.join(p, "Save")
        #         print("found 'local' save dir ", repr(cls.save_dir))
        #         break
        #     path = p
        #     p = os.path.dirname(path)
        # if len(cls.game_dirs) == 0:
        #     while True:
        #         path = Config.get("GameDirectory/%d" % k)
        #         if not path:
        #             break
        #         cls.game_dirs.append(path.strip())
        #         k += 1
        # if len(cls.game_dirs) == 0:
        #     path = os.path.join(get_home_dir(), "Games")
        #     if os.path.exists(path):
        #         print("auto add", path)
        #         cls.game_dirs.append(path)
        #     path = os.path.join(get_documents_dir(), "Games")
        #     if os.path.exists(path):
        #         print("auto add", path)
        #         cls.game_dirs.append(path)
        #     if windows:
        #         path = "C:\\Games"
        #         if os.path.exists(path):
        #             print("auto add", path)
        #             cls.game_dirs.append(path)
        # print("game dirs:", cls.game_dirs)
        #
        # if not cls.save_dir:
        #     # FIXME: Use known folder SAVED_GAMES (or something) on Vista+
        #     cls.save_dir = os.path.join(get_home_dir(), "Games", "Saves")
        #
        # if not cls.data_dir:
        #     cls.data_dir = fs.get_app_data_dir()

    @classmethod
    def get_local_data_dir(cls):
        # """local data dir with resources overriding those downloaded from
        # the game metadata server."""
        # data_dir = fs.get_app_data_dir()
        # data_dir = data_dir.replace(fs.get_app_id(), "fs-game-database")
        # data_dir = os.path.join(data_dir, "data")
        # return data_dir
        raise NotImplementedError()

    @classmethod
    def register_user_activity(cls):
        if windows:
            pass
        elif macosx:
            pass
        else:
            path = "/tmp/fs-game-center-activity"
            if not os.path.exists(path):
                try:
                    with open(path, "wb") as _:
                        pass
                except Exception:
                    pass
                try:
                    os.chmod(path, 0o666)
                except Exception:
                    pass
            os.utime("/tmp/fs-game-center-activity", None)



from fsgs.context import fsgs
import fsui as fsui
from fsgs.amiga.Amiga import Amiga
from ..Signal import Signal
from ..I18N import gettext


class KickstartStatusGroup(fsui.Group):

    def __init__(self, parent, title, model):
        self.model = model
        fsui.Group.__init__(self, parent)
        self.layout = fsui.HorizontalLayout()

        self.ok_image = fsui.Image("fs_uae_launcher:res/ok_emblem.png")
        self.na_image = fsui.Image("fs_uae_launcher:res/na_emblem.png")

        self.icon = fsui.ImageView(self, self.na_image)
        self.layout.add(self.icon)

        self.layout.add_spacer(10)
        self.label = fsui.Label(self, title)
        self.layout.add(self.label)
        self.update()
        Signal.add_listener("scan_done", self)

    def on_destroy(self):
        Signal.remove_listener("scan_done", self)

    def on_scan_done_signal(self):
        self.update()

    def update(self):
        amiga = Amiga.get_model_config(self.model)
        for sha1 in amiga["kickstarts"]:
            if fsgs.file.find_by_sha1(sha1):
                self.icon.set_image(self.ok_image)
                return
        self.icon.set_image(self.na_image)


class ScanKickstartGroup(fsui.Group):

    def __init__(self, parent):
        fsui.Group.__init__(self, parent)

        self.layout = fsui.VerticalLayout()
        label = fsui.HeadingLabel(
            self, gettext("Available Kickstart Versions"))
        self.layout.add(label, margin_bottom=10)

        icon_layout = fsui.HorizontalLayout()
        self.layout.add(icon_layout, fill=True)

        icon_layout.add_spacer(20)
        image = fsui.Image("fs_uae_launcher:res/kickstart.png")
        self.image_view = fsui.ImageView(self, image)
        icon_layout.add(self.image_view, valign=0.0, margin_right=10)

        vert_layout = fsui.VerticalLayout()
        icon_layout.add(vert_layout, fill=True, expand=True)

        vert_layout.add_spacer(0)

        label = fsui.Label(
            self, gettext("You should have kickstart files for "
                          "each Amiga model you want to use:"))
        vert_layout.add(label, margin_bottom=0)

        hori_layout = fsui.HorizontalLayout()
        vert_layout.add(hori_layout, fill=True)

        self.kickstart_groups = []

        column_layout = fsui.VerticalLayout()
        hori_layout.add(column_layout, expand=True, fill=True, margin=10)

        self.add_kickstart_group(column_layout, "Amiga 1000", "A1000")
        column_layout.add_spacer(10)
        self.add_kickstart_group(column_layout, "Amiga 500", "A500")
        column_layout.add_spacer(10)
        self.add_kickstart_group(column_layout, "Amiga 500+", "A500+")

        column_layout = fsui.VerticalLayout()
        hori_layout.add(column_layout, expand=True, fill=True, margin=10)

        self.add_kickstart_group(column_layout, "Amiga 600", "A600")
        column_layout.add_spacer(10)
        self.add_kickstart_group(column_layout, "Amiga 1200", "A1200")
        column_layout.add_spacer(10)
        self.add_kickstart_group(column_layout, "Amiga 3000", "A3000")

        column_layout = fsui.VerticalLayout()
        hori_layout.add(column_layout, expand=True, fill=True, margin=10)

        self.add_kickstart_group(column_layout, "Amiga 4000", "A4000/040")
        column_layout.add_spacer(10)
        self.add_kickstart_group(column_layout, "Amiga CD32", "CD32")
        column_layout.add_spacer(10)
        self.add_kickstart_group(column_layout, "Commodore CDTV", "CDTV")

    def add_kickstart_group(self, layout, title, model):
        group = KickstartStatusGroup(self, title, model)
        self.kickstart_groups.append(group)
        layout.add(group, fill=True)



import os
from xml.etree.cElementTree import ElementTree, SubElement
from xml.etree.cElementTree import fromstring, tostring
import fsui as fsui


class XMLControl(fsui.TextArea):

    def __init__(self, parent):
        fsui.TextArea.__init__(self, parent)  # , horizontal_scroll=True)
        self.path = ""

    def connect_game(self, info):
        tree = self.get_tree()
        root = tree.getroot()
        if not root.tag == "config":
            return
        game_node = self.find_or_create_node(root, "game")
        game_node.set("uuid", info["uuid"])
        game_name_node = self.find_or_create_node(game_node, "name")
        game_name_node.text = info["name"]
        self.set_tree(tree)

    def find_or_create_node(self, element, name):
        node = element.find(name)
        if node is None:
            node = SubElement(element, name)
        return node

    def set_path(self, path):
        if not os.path.exists(path):
            path = ""
        self.path = path
        if path:
            self.load_xml(path)
        else:
            self.set_text("")

    def get_tree(self):
        text = self.get_text().strip()
        try:
            root = fromstring(text.encode("UTF-8"))
        except Exception:
            # FIXME: show message
            import traceback
            traceback.print_exc()
            return
        tree = ElementTree(root)
        indent_tree(root)
        return tree

    def set_tree(self, tree):
        data = tostring(tree.getroot(), encoding="UTF-8").decode("UTF-8")
        std_decl = "<?xml version='1.0' encoding='UTF-8'?>"
        if data.startswith(std_decl):
            data = data[len(std_decl):].strip()
        self.set_text(data)

    def load_xml(self, path):
        with open(path, "rb") as f:
            data = f.read()
            self.set_text(data)

    def save(self):
        if not self.path:
            print("no path to save XML to")
            return
        self.save_xml(self.path)

    def save_xml(self, path):
        self.get_tree().write(self.path)


def indent_tree(elem, level=0):
    i = "\n" + level * "  "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for elem in elem:
            indent_tree(elem, level+1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i



from .opengl import *
from .render import Render
from .window import set_program
from .window import texture_program, premultiplied_texture_program


class FrameBufferObject(object):  # Renderable, RendererBase):

    def __init__(self, width=100, height=100):
        self._transparent = True
        self._has_transparency = True
        self._width = width
        self._height = height
        self._fb = None
        self._depth_rb = None
        self._stencil_rb = None
        self._texture = None

    def __del__(self):
        self.free()

    def free(self):
        if self._fb:
            glDeleteFramebuffersEXT(1, self._fb)
        if self._depth_rb:
            glDeleteRenderbuffersEXT(1, self._depth_rb)
        if self._stencil_rb:
            glDeleteRenderbuffersEXT(1, self._stencil_rb)
        if self._texture:
            glDeleteTextures([self._texture])

    def _create(self):
        if self._fb is not None:
            return

        w = self._width
        h = self._height
        if self._transparent:
            texture_format = GL_RGBA
        else:
            texture_format = GL_RGBA
        texture_target = GL_TEXTURE_2D
        fb = GLuint()
        depth_rb = GLuint()
        stencil_rb = GLuint()

        glGenFramebuffersEXT(1, ctypes.byref(fb))
        color_tex = Render.create_texture()
        glGenRenderbuffersEXT(1, ctypes.byref(depth_rb))
        glGenRenderbuffersEXT(1, ctypes.byref(stencil_rb))

        fb = fb.value
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb)

        # initialize color texture
        glBindTexture(texture_target, color_tex)
        glTexParameterf(texture_target, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
        glTexParameterf(texture_target, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        glTexImage2D(
            texture_target, 0, texture_format, w, h, 0, GL_RGB, GL_INT, None)
        try:
            # FIXME: An error seems to be thrown here on Windows, even
            # though it seems to work...
            glFramebufferTexture2DEXT(
                GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, texture_target,
                color_tex, 0)
        except Exception:
            pass

        # initialize depth renderbuffer
        glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, depth_rb)
        glRenderbufferStorageEXT(
            GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT24, w, h)
        glFramebufferRenderbufferEXT(
            GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT,
            depth_rb)

        # initialize stencil renderbuffer

        # glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, stencil_rb)
        # glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_STENCIL_INDEX,
        #         512, 512)
        # glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,
        #        GL_STENCIL_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, stencil_rb)
        glBindTexture(texture_target, 0)

        # Check framebuffer completeness at the end of initialization.
        status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT)
        assert status == GL_FRAMEBUFFER_COMPLETE_EXT, str(status)
        self._fb = fb
        self._depth_rb = depth_rb
        self._stencil_rb = stencil_rb
        self._texture = color_tex

    def bind(self):
        if self._fb is None:
            self._create()
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, self._fb)
        # glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, self._depth_rb)

    def unbind(self):
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0)
        # glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, 0)

    def set_viewport(self):
        glViewport(0, 0, self._width, self._height)

    def render(self, opacity=1.0):
        render_texture(
            self._texture, -1.0, -1.0, 0, 2.0, 2.0, opacity=opacity,
            is_premultiplied=True)


def render_texture(
        texture, x, y, z, w, h, opacity=1.0, is_premultiplied=False):
    if is_premultiplied:
        set_program(premultiplied_texture_program)
    else:
        set_program(texture_program)
    fs_emu_texturing(True)
    glBindTexture(GL_TEXTURE_2D, texture)
    glBegin(GL_QUADS)
    glColor(1.0, 1.0, 1.0, opacity)
    glTexCoord2f(0.0, 0.0)
    glVertex3f(x, y, z)
    glTexCoord2f(1.0, 0.0)
    glVertex3f(x + w, y, z)
    glTexCoord2f(1.0, 1.0)
    glVertex3f(x + w, y + h, z)
    glTexCoord2f(0.0, 1.0)
    glVertex3f(x, y + h, z)
    glEnd()
    set_program(None)



# FSGS - Common functionality for Fengestad Game System.
# Copyright (C) 2013  Frode Solheim <frode-code@fengestad.no>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Game Runner for DOS.

"""
import os
import io
import json
import hashlib
from fsbc.path import str_path
from fsbc.system import windows
from fsgs.runner import GameRunner


# noinspection PyAttributeOutsideInit
class DOSRunner(GameRunner):

    def __init__(self, fsgs):
        super().__init__(fsgs)

    def prepare(self):
        self.temp_dir = self.create_temp_dir("dosbox")
        self.drives_dir = self.create_temp_dir("dosbox-drives")
        config_file = os.path.join(self.temp_dir.path, "dosbox.cfg")
        self.drives = []
        self.prepare_media()
        with io.open(config_file, "w", encoding="UTF-8") as f:
            self.configure(f)
        self.add_arg("-conf", config_file)

    def prepare_media(self):
        file_list = json.loads(self.config["file_list"])
        self.unpack_game_hard_drives(file_list)

    def unpack_game_hard_drives(self, file_list):
        # dir_path = os.path.join(self.temp_dir, dir_name)
        drives_added = set()
        dir_path = self.drives_dir.path
        for file_entry in file_list:
            # if self.stop_flag:
            #     return
            name = file_entry["name"].upper()
            drives_added.add(name[0])

            # extract relative path and convert each path component
            # to host file name (where needed).

            rel_path = name
            print("rel_path", rel_path)
            rel_parts = rel_path.split("/")
            # for i, part in enumerate(rel_parts):
            #     # part can be blank if rel_parts is a directory
            #     # (ending with /)
            #     if part:
            #         rel_parts[i] = amiga_filename_to_host_filename(part)
            rel_path = "/".join(rel_parts)

            dst_file = os.path.join(dir_path, rel_path)
            print(repr(dst_file))
            if name.endswith("/"):
                os.makedirs(str_path(dst_file))
                continue
            sha1 = file_entry["sha1"]
            src_file = self.fsgs.file.find_by_sha1(sha1)
            if not os.path.exists(os.path.dirname(dst_file)):
                os.makedirs(os.path.dirname(dst_file))
            stream = self.fsgs.file.open(src_file)
            # archive = Archive(src_file)
            # f = archive.open(src_file)
            data = stream.read()
            assert hashlib.sha1(data).hexdigest() == sha1

            with open(dst_file, "wb") as out_file:
                out_file.write(data)

        for drive in sorted(drives_added):
            self.drives.append((drive, os.path.join(dir_path, drive)))

    def configure(self, f):
        f.write("# automatically generated\n")
        f.write("\n[sdl]\n")
        f.write("output=opengl\n")

        if self.use_fullscreen():
            f.write("fullscreen=true\n")
        else:
            f.write("fullscreen=false\n")
        f.write("fullresolution={0}x{1}\n".format(*self.screen_size()))

        f.write("\n[render]\n")
        f.write("frameskip=0\n")
        if self.use_stretching():
            # This option does not stretch, it merely does not correct
            # aspect for non-square pixels resolutions, e.g. 320x200
            f.write("aspect=false\n")
            # This custom environment variable however, does cause stretching
            self.set_env("FSGS_STRETCH", "1")
        else:
            f.write("aspect=true\n")
            self.set_env("FSGS_STRETCH", "0")

        f.write("\n[cpu]\n")
        cpu_core = "auto"
        f.write("core={0}\n".format(cpu_core))
        cpu_cycles = "auto"
        f.write("cycles={0}\n".format(cpu_cycles))

        f.write("\n[autoexec]\n")
        f.write("@echo off\n")
        # for drive, drive_path in self.drives:
        #     pass
        for name in os.listdir(self.drives_dir.path):
            if name in "ABCDEFGHIJKLMNOPQRSTUVWXYZ":
                if name in "DEF":
                    f.write('MOUNT {0} "{1}" -t cdrom\n'.format(name,
                            os.path.join(self.drives_dir.path, name)))
                else:
                    f.write('MOUNT {0} "{1}"\n'.format(name,
                            os.path.join(self.drives_dir.path, name)))
        f.write("C:\n")
        f.write("CLS\n")
        # for i in range(25):
        #     f.write("echo.\n")
        for command in self.config["hd_startup"].split(";"):
            command = command.strip()
            f.write("{0}\n".format(command))
        f.write("exit\n")

        if windows:
            # we don't want to open the separate console window on windows
            self.add_arg("-noconsole")

    def run(self):
        return self.start_emulator_from_plugin_resource("fs-dosbox")

    def finish(self):
        pass



'''OpenGL extension VERSION.GL_3_0

This module customises the behaviour of the 
OpenGL.raw.GL.VERSION.GL_3_0 to provide a more 
Python-friendly API

The official definition of this extension is available here:
http://www.opengl.org/registry/specs/VERSION/GL_3_0.txt
'''
from OpenGL import platform, constant, arrays
from OpenGL import extensions, wrapper
from OpenGL.GL import glget
import ctypes
from OpenGL.raw.GL import _types
from OpenGL.raw.GL.VERSION.GL_3_0 import *
from OpenGL.raw.GL.VERSION.GL_3_0 import _EXTENSION_NAME

def glInitGl30VERSION():
    '''Return boolean indicating whether this extension is available'''
    from OpenGL import extensions
    return extensions.hasGLExtension( _EXTENSION_NAME )

### END AUTOGENERATED SECTION
glget.addGLGetConstant( GL_NUM_EXTENSIONS, (1,))
from ctypes import c_char_p
glGetStringi.restype = c_char_p

from OpenGL.GL.ARB.vertex_array_object import *
from OpenGL.GL.ARB.texture_buffer_object import *
from OpenGL.GL.ARB.framebuffer_object import *
from OpenGL.GL.ARB.map_buffer_range import *

GL_DEPTH_BUFFER = constant.Constant( 'GL_DEPTH_BUFFER', 0x8223 )
GL_STENCIL_BUFFER = constant.Constant( 'GL_STENCIL_BUFFER', 0x8224 )

glGenVertexArrays = wrapper.wrapper(glGenVertexArrays).setOutput(
    'arrays', lambda n: (n,), 'n', arrayType = arrays.GLuintArray,
    orPassIn=True,
)



from fsui.qt import QPushButton, QSignal
from .Widget import Widget


class Button(QPushButton, Widget):

    activated = QSignal()

    def __init__(self, parent, label):
        label = "  " + label + "  "
        # self._widget = QPushButton(label, parent.get_container())
        QPushButton.__init__(self, label, parent.get_container())
        # Widget.__init__(self, parent)
        self.init_widget(parent)
        # self._widget.clicked.connect(self.__clicked)
        # if not System.macosx:
        #     self.set_min_height(28)
        self.clicked.connect(self.__clicked)

    def __clicked(self):
        self.on_activate()
        self.activated.emit()

    def on_activate(self):
        pass



"""Image-handling routines

### Unresolved:

    Following methods are not yet resolved due to my not being sure how the
    function should be wrapped:

        glCompressedTexImage3D
        glCompressedTexImage2D
        glCompressedTexImage1D
        glCompressedTexSubImage3D
        glCompressedTexSubImage2D
        glCompressedTexSubImage1D
"""
from OpenGL.raw.GL.VERSION import GL_1_1,GL_1_2, GL_3_0
from OpenGL import images, arrays, wrapper
from OpenGL._bytes import bytes,integer_types
import ctypes

def asInt( value ):
    if isinstance( value, float ):
        return int(round(value,0))
    return value

## update the image tables with standard image types...
images.COMPONENT_COUNTS.update( {
    GL_1_1.GL_BITMAP : 1, # must be GL_UNSIGNED_BYTE

    GL_1_1.GL_RED : 1,
    GL_1_1.GL_GREEN : 1,
    GL_1_1.GL_BLUE : 1,
    GL_1_1.GL_ALPHA : 1,
    GL_3_0.GL_RED_INTEGER : 1,
    GL_3_0.GL_GREEN_INTEGER : 1,
    GL_3_0.GL_BLUE_INTEGER : 1,
    GL_3_0.GL_ALPHA_INTEGER : 1,
    GL_1_1.GL_LUMINANCE : 1,
    GL_1_1.GL_LUMINANCE_ALPHA : 2,
    GL_1_1.GL_COLOR_INDEX : 1,
    GL_1_1.GL_STENCIL_INDEX : 1,
    GL_1_1.GL_DEPTH_COMPONENT : 1,

    GL_1_1.GL_RGB : 3,
    GL_1_2.GL_BGR : 3,
    GL_3_0.GL_RGB16F : 3,
    GL_3_0.GL_RGB16I : 3,
    GL_3_0.GL_RGB16UI : 3,
    GL_3_0.GL_RGB32F : 3,
    GL_3_0.GL_RGB32I : 3,
    GL_3_0.GL_RGB32UI : 3,
    GL_3_0.GL_RGB8I : 3,
    GL_3_0.GL_RGB8UI : 3,
    GL_3_0.GL_RGB9_E5 : 3,
    GL_3_0.GL_RGB_INTEGER : 3,

    GL_1_1.GL_RGBA : 4,
    GL_1_2.GL_BGRA : 4,
    GL_3_0.GL_RGBA16F : 4,
    GL_3_0.GL_RGBA16I : 4,
    GL_3_0.GL_RGBA16UI : 4,
    GL_3_0.GL_RGBA32F : 4,
    GL_3_0.GL_RGBA32I : 4,
    GL_3_0.GL_RGBA32UI : 4,
    GL_3_0.GL_RGBA8I : 4,
    GL_3_0.GL_RGBA8UI : 4,
    GL_3_0.GL_RGBA_INTEGER : 4,
} )

images.TYPE_TO_ARRAYTYPE.update( {
    GL_1_2.GL_UNSIGNED_BYTE_3_3_2 : GL_1_1.GL_UNSIGNED_BYTE,
    GL_1_2.GL_UNSIGNED_BYTE_2_3_3_REV : GL_1_1.GL_UNSIGNED_BYTE,
    GL_1_2.GL_UNSIGNED_SHORT_4_4_4_4 : GL_1_1.GL_UNSIGNED_SHORT,
    GL_1_2.GL_UNSIGNED_SHORT_4_4_4_4_REV : GL_1_1.GL_UNSIGNED_SHORT,
    GL_1_2.GL_UNSIGNED_SHORT_5_5_5_1 : GL_1_1.GL_UNSIGNED_SHORT,
    GL_1_2.GL_UNSIGNED_SHORT_1_5_5_5_REV : GL_1_1.GL_UNSIGNED_SHORT,
    GL_1_2.GL_UNSIGNED_SHORT_5_6_5 : GL_1_1.GL_UNSIGNED_SHORT,
    GL_1_2.GL_UNSIGNED_SHORT_5_6_5_REV : GL_1_1.GL_UNSIGNED_SHORT,
    GL_1_2.GL_UNSIGNED_INT_8_8_8_8 : GL_1_1.GL_UNSIGNED_INT,
    GL_1_2.GL_UNSIGNED_INT_8_8_8_8_REV : GL_1_1.GL_UNSIGNED_INT,
    GL_1_2.GL_UNSIGNED_INT_10_10_10_2 : GL_1_1.GL_UNSIGNED_INT,
    GL_1_2.GL_UNSIGNED_INT_2_10_10_10_REV : GL_1_1.GL_UNSIGNED_INT,
    GL_1_1.GL_UNSIGNED_BYTE : GL_1_1.GL_UNSIGNED_BYTE,
    GL_1_1.GL_BYTE: GL_1_1.GL_BYTE,
    GL_1_1.GL_UNSIGNED_SHORT : GL_1_1.GL_UNSIGNED_SHORT,
    GL_1_1.GL_SHORT :  GL_1_1.GL_SHORT,
    GL_1_1.GL_UNSIGNED_INT : GL_1_1.GL_UNSIGNED_INT,
    GL_1_1.GL_INT : GL_1_1.GL_INT,
    GL_1_1.GL_FLOAT : GL_1_1.GL_FLOAT,
    GL_1_1.GL_DOUBLE : GL_1_1.GL_DOUBLE,
    GL_1_1.GL_BITMAP : GL_1_1.GL_UNSIGNED_BYTE,
} )
images.TIGHT_PACK_FORMATS.update({
    GL_1_2.GL_UNSIGNED_BYTE_3_3_2 : 3,
    GL_1_2.GL_UNSIGNED_BYTE_2_3_3_REV : 3,
    GL_1_2.GL_UNSIGNED_SHORT_4_4_4_4 : 4,
    GL_1_2.GL_UNSIGNED_SHORT_4_4_4_4_REV : 4,
    GL_1_2.GL_UNSIGNED_SHORT_5_5_5_1 : 4,
    GL_1_2.GL_UNSIGNED_SHORT_1_5_5_5_REV : 4,
    GL_1_2.GL_UNSIGNED_SHORT_5_6_5 : 3,
    GL_1_2.GL_UNSIGNED_SHORT_5_6_5_REV : 3,
    GL_1_2.GL_UNSIGNED_INT_8_8_8_8 : 4,
    GL_1_2.GL_UNSIGNED_INT_8_8_8_8_REV : 4,
    GL_1_2.GL_UNSIGNED_INT_10_10_10_2 : 4,
    GL_1_2.GL_UNSIGNED_INT_2_10_10_10_REV : 4,
    GL_1_1.GL_BITMAP: 8, # single bits, 8 of them...
})

images.RANK_PACKINGS.update( {
    4: [
        # Note the sgis parameters are skipped here unless you import 
        # the sgis texture4D extension...
        (GL_1_1.glPixelStorei,GL_1_1.GL_PACK_ALIGNMENT, 1),
    ],
    3: [
        (GL_1_1.glPixelStorei,GL_1_2.GL_PACK_SKIP_IMAGES, 0),
        (GL_1_1.glPixelStorei,GL_1_2.GL_PACK_IMAGE_HEIGHT, 0),
        (GL_1_1.glPixelStorei,GL_1_1.GL_PACK_ALIGNMENT, 1),
    ],
    2: [
        (GL_1_1.glPixelStorei,GL_1_1.GL_PACK_ROW_LENGTH, 0),
        (GL_1_1.glPixelStorei,GL_1_1.GL_PACK_SKIP_ROWS, 0),
        (GL_1_1.glPixelStorei,GL_1_1.GL_PACK_ALIGNMENT, 1),
    ],
    1: [
        (GL_1_1.glPixelStorei,GL_1_1.GL_PACK_SKIP_PIXELS, 0),
        (GL_1_1.glPixelStorei,GL_1_1.GL_PACK_ALIGNMENT, 1),
    ],
} )


__all__ = (
    'glReadPixels',
    'glReadPixelsb',
    'glReadPixelsd',
    'glReadPixelsf',
    'glReadPixelsi',
    'glReadPixelss',
    'glReadPixelsub',
    'glReadPixelsui',
    'glReadPixelsus',

    'glGetTexImage',

    'glDrawPixels',
    'glDrawPixelsb',
    'glDrawPixelsf',
    'glDrawPixelsi',
    'glDrawPixelss',
    'glDrawPixelsub',
    'glDrawPixelsui',
    'glDrawPixelsus',


    'glTexSubImage2D',
    'glTexSubImage1D',
    #'glTexSubImage3D',

    'glTexImage1D',
    'glTexImage2D',
    #'glTexImage3D',

    'glGetTexImageb',
    'glGetTexImaged',
    'glGetTexImagef',
    'glGetTexImagei',
    'glGetTexImages',
    'glGetTexImageub',
    'glGetTexImageui',
    'glGetTexImageus',
    'glTexImage1Db',
    'glTexImage2Db',
    #'glTexImage3Db',
    'glTexSubImage1Db',
    'glTexSubImage2Db',
    #'glTexSubImage3Db',
    'glTexImage1Df',
    'glTexImage2Df',
    #'glTexImage3Df',
    'glTexSubImage1Df',
    'glTexSubImage2Df',
    #'glTexSubImage3Df',
    'glTexImage1Di',
    'glTexImage2Di',
    #'glTexImage3Di',
    'glTexSubImage1Di',
    'glTexSubImage2Di',
    #'glTexSubImage3Di',
    'glTexImage1Ds',
    'glTexImage2Ds',
    #'glTexImage3Ds',
    'glTexSubImage1Ds',
    'glTexSubImage2Ds',
    #'glTexSubImage3Ds',
    'glTexImage1Dub',
    'glTexImage2Dub',
    #'glTexImage3Dub',
    'glTexSubImage1Dub',
    'glTexSubImage2Dub',
    #'glTexSubImage3Dub',
    'glTexImage1Dui',
    'glTexImage2Dui',
    #'glTexImage3Dui',
    'glTexSubImage1Dui',
    'glTexSubImage2Dui',
    #'glTexSubImage3Dui',
    'glTexImage1Dus',
    'glTexImage2Dus',
    #'glTexImage3Dus',
    'glTexSubImage1Dus',
    'glTexSubImage2Dus',
    #'glTexSubImage3Dus',

    #'glColorTable',
    #'glGetColorTable',
    #'glColorSubTable',

    #'glConvolutionFilter1D',
    #'glConvolutionFilter2D',
    #'glGetConvolutionFilter',
    #'glSeparableFilter2D',
    #'glGetSeparableFilter',

    #'glGetMinmax',
)

for suffix,type in [
    ('b',GL_1_1.GL_BYTE),
    ('d',GL_1_1.GL_DOUBLE),
    ('f',GL_1_1.GL_FLOAT),
    ('i',GL_1_1.GL_INT),
    ('s',GL_1_1.GL_SHORT),
    ('ub',GL_1_1.GL_UNSIGNED_BYTE),
    ('ui',GL_1_1.GL_UNSIGNED_INT),
    ('us',GL_1_1.GL_UNSIGNED_SHORT),
]:
    def glReadPixels( x,y,width,height,format,type=type, array=None, outputType=bytes ):
        """Read specified pixels from the current display buffer

        This typed version returns data in your specified default
        array data-type format, or in the passed array, which will
        be converted to the array-type required by the format.
        """
        x,y,width,height = asInt(x),asInt(y),asInt(width),asInt(height)
        arrayType = arrays.GL_CONSTANT_TO_ARRAY_TYPE[ images.TYPE_TO_ARRAYTYPE.get(type,type) ]
        
        if array is None:
            array = imageData = images.SetupPixelRead( format, (width,height), type )
            owned = True
        else:
            if isinstance( array, integer_types):
                imageData = ctypes.c_void_p( array )
            else:
                array = arrayType.asArray( array )
                imageData = arrayType.voidDataPointer( array )
            owned = False
        GL_1_1.glReadPixels(
            x,y,
            width, height,
            format,type,
            imageData
        )
        if owned and outputType is bytes:
            return images.returnFormat( array, type )
        else:
            return array
    globals()["glReadPixels%s"%(suffix,)] = glReadPixels
    def glGetTexImage( target, level,format,type=type, array=None, outputType=bytes ):
        """Get a texture-level as an image
        
        target -- enum constant for the texture engine to be read
        level -- the mip-map level to read
        format -- image format to read out the data
        type -- data-type into which to read the data
        array -- optional array/offset into which to store the value

        outputType -- default (bytes) provides string output of the
            results iff OpenGL.UNSIGNED_BYTE_IMAGES_AS_STRING is True
            and type == GL_UNSIGNED_BYTE.  Any other value will cause
            output in the default array output format.

        returns the pixel data array in the format defined by the
        format, type and outputType
        """
        from OpenGL.GL import glget
        arrayType = arrays.GL_CONSTANT_TO_ARRAY_TYPE[ images.TYPE_TO_ARRAYTYPE.get(type,type) ]
        if array is None:
            dims = [glget.glGetTexLevelParameteriv( target, level, GL_1_1.GL_TEXTURE_WIDTH )]
            if target != GL_1_1.GL_TEXTURE_1D:
                dims.append( glget.glGetTexLevelParameteriv( target, level, GL_1_1.GL_TEXTURE_HEIGHT ) )
                if target != GL_1_1.GL_TEXTURE_2D:
                    dims.append( glget.glGetTexLevelParameteriv( target, level, GL_1_2.GL_TEXTURE_DEPTH ) )
            array = imageData = images.SetupPixelRead( format, tuple(dims), type )
            owned = True
        else:
            if isinstance( array, integer_types):
                imageData = ctypes.c_void_p( array )
            else:
                array = arrayType.asArray( array )
                imageData = arrayType.voidDataPointer( array )
            owned = False
        GL_1_1.glGetTexImage(
            target, level, format, type, imageData
        )
        if owned and outputType is bytes:
            return images.returnFormat( array, type )
        else:
            return array
    globals()["glGetTexImage%s"%(suffix,)] = glGetTexImage
##	def glGetTexSubImage( target, level,format,type ):
##		"""Get a texture-level as an image"""
##		from OpenGL.GL import glget
##		dims = [glget.glGetTexLevelParameteriv( target, level, GL_1_1.GL_TEXTURE_WIDTH )]
##		if target != GL_1_1.GL_TEXTURE_1D:
##			dims.append( glget.glGetTexLevelParameteriv( target, level, GL_1_1.GL_TEXTURE_HEIGHT ) )
##			if target != GL_1_1.GL_TEXTURE_2D:
##				dims.append( glget.glGetTexLevelParameteriv( target, level, GL_1_2.GL_TEXTURE_DEPTH ) )
##		array = images.SetupPixelRead( format, tuple(dims), type )
##		arrayType = arrays.GL_CONSTANT_TO_ARRAY_TYPE[ images.TYPE_TO_ARRAYTYPE.get(type,type) ]
##		GL_1_1.glGetTexImage(
##			target, level, format, type, ctypes.c_void_p( arrayType.dataPointer(array))
##		)
##		return array
##	"%s = glGetTexImage"%(suffix)
    try:
        del suffix,type
    except NameError as err:
        pass
# Now the real glReadPixels...
def glReadPixels( x,y,width,height,format,type, array=None, outputType=bytes ):
    """Read specified pixels from the current display buffer

    x,y,width,height -- location and dimensions of the image to read
        from the buffer
    format -- pixel format for the resulting data
    type -- data-format for the resulting data
    array -- optional array/offset into which to store the value
    outputType -- default (bytes) provides string output of the
        results iff OpenGL.UNSIGNED_BYTE_IMAGES_AS_STRING is True
        and type == GL_UNSIGNED_BYTE.  Any other value will cause
        output in the default array output format.

    returns the pixel data array in the format defined by the
    format, type and outputType
    """
    x,y,width,height = asInt(x),asInt(y),asInt(width),asInt(height)

    arrayType = arrays.GL_CONSTANT_TO_ARRAY_TYPE[ images.TYPE_TO_ARRAYTYPE.get(type,type) ]
    if array is None:
        array = imageData = images.SetupPixelRead( format, (width,height), type )
        owned = True
    else:
        if isinstance( array, integer_types):
            imageData = ctypes.c_void_p( array )
        else:
            array = arrayType.asArray( array )
            imageData = arrayType.voidDataPointer( array )
        owned = False

    GL_1_1.glReadPixels(
        x,y,width,height,
        format,type,
        imageData
    )
    if owned and outputType is bytes:
        return images.returnFormat( array, type )
    else:
        return array

def glGetTexImage( target, level,format,type, array=None, outputType=bytes ):
    """Get a texture-level as an image

    target -- enum constant for the texture engine to be read
    level -- the mip-map level to read
    format -- image format to read out the data
    type -- data-type into which to read the data
    array -- optional array/offset into which to store the value

    outputType -- default (bytes) provides string output of the
        results iff OpenGL.UNSIGNED_BYTE_IMAGES_AS_STRING is True
        and type == GL_UNSIGNED_BYTE.  Any other value will cause
        output in the default array output format.

    returns the pixel data array in the format defined by the
    format, type and outputType
    """
    from OpenGL.GL import glget
    arrayType = arrays.GL_CONSTANT_TO_ARRAY_TYPE[ images.TYPE_TO_ARRAYTYPE.get(type,type) ]
    if array is None:
        dims = [glget.glGetTexLevelParameteriv( target, level, GL_1_1.GL_TEXTURE_WIDTH )]
        if target != GL_1_1.GL_TEXTURE_1D:
            dims.append( glget.glGetTexLevelParameteriv( target, level, GL_1_1.GL_TEXTURE_HEIGHT ) )
            if target != GL_1_1.GL_TEXTURE_2D:
                dims.append( glget.glGetTexLevelParameteriv( target, level, GL_1_2.GL_TEXTURE_DEPTH ) )
        array = imageData = images.SetupPixelRead( format, tuple(dims), type )
        owned = True
    else:
        if isinstance( array, integer_types):
            imageData = ctypes.c_void_p( array )
        else:
            array = arrayType.asArray( array )
            imageData = arrayType.voidDataPointer( array )
        owned = False
    GL_1_1.glGetTexImage(
        target, level, format, type, imageData
    )
    if outputType is bytes:
        return images.returnFormat( array, type )
    else:
        return array


INT_DIMENSION_NAMES = [
    'width','height','depth','x','y','z',
    'xoffset','yoffset','zoffset',
    'start', 'count',
]
def asWrapper( value ):
    if not isinstance( value, wrapper.Wrapper ):
        return wrapper.wrapper( value )
    return value

def asIntConverter( value, *args ):
    if isinstance( value, float ):
        return int(round(value,0))
    return value

def setDimensionsAsInts( baseOperation ):
    """Set arguments with names in INT_DIMENSION_NAMES to asInt processing"""
    baseOperation = asWrapper( baseOperation )
    argNames = getattr( baseOperation, 'pyConverterNames', baseOperation.argNames )
    for i,argName in enumerate(argNames):
        if argName in INT_DIMENSION_NAMES:
            baseOperation.setPyConverter( argName, asIntConverter )
    return baseOperation



class ImageInputConverter( object ):
    def __init__( self, rank, pixelsName=None, typeName='type' ):
        self.rank = rank
        self.typeName = typeName
        self.pixelsName = pixelsName
    def finalise( self, wrapper ):
        """Get our pixel index from the wrapper"""
        self.typeIndex = wrapper.pyArgIndex( self.typeName )
        self.pixelsIndex = wrapper.pyArgIndex( self.pixelsName )
    def __call__( self, arg, baseOperation, pyArgs ):
        """pyConverter for the pixels argument"""
        images.setupDefaultTransferMode()
        images.rankPacking( self.rank )
        type = pyArgs[ self.typeIndex ]
        arrayType = arrays.GL_CONSTANT_TO_ARRAY_TYPE[ images.TYPE_TO_ARRAYTYPE[ type ] ]
        return arrayType.asArray( arg )
#	def cResolver( self, array ):
#		return array
#		return ctypes.c_void_p( arrays.ArrayDatatype.dataPointer( array ) )

class TypedImageInputConverter( ImageInputConverter ):
    def __init__( self, rank, pixelsName, arrayType, typeName=None ):
        self.rank = rank
        self.arrayType = arrayType
        self.pixelsName = pixelsName
        self.typeName = typeName
    def __call__( self, arg, baseOperation, pyArgs ):
        """The pyConverter for the pixels"""
        images.setupDefaultTransferMode()
        images.rankPacking( self.rank )
        return self.arrayType.asArray( arg )
    def finalise( self, wrapper ):
        """Get our pixel index from the wrapper"""
        self.pixelsIndex = wrapper.pyArgIndex( self.pixelsName )
    def width( self, pyArgs, index, wrappedOperation ):
        """Extract the width from the pixels argument"""
        return self.arrayType.dimensions( pyArgs[self.pixelsIndex] )[0]
    def height( self, pyArgs, index, wrappedOperation ):
        """Extract the height from the pixels argument"""
        return self.arrayType.dimensions( pyArgs[self.pixelsIndex] )[1]
    def depth( self, pyArgs, index, wrappedOperation ):
        """Extract the depth from the pixels argument"""
        return self.arrayType.dimensions( pyArgs[self.pixelsIndex] )[2]
    def type( self, pyArgs, index, wrappedOperation ):
        """Provide the item-type argument from our stored value

        This is used for pre-bound processing where we want to provide
        the type by implication...
        """
        return self.typeName

class CompressedImageConverter( object ):
    def finalise( self, wrapper ):
        """Get our pixel index from the wrapper"""
        self.dataIndex = wrapper.pyArgIndex( 'data' )
    def __call__( self, pyArgs, index, wrappedOperation ):
        """Create a data-size measurement for our image"""
        arg = pyArgs[ self.dataIndex ]
        return arrays.ArrayType.arrayByteCount( arg )



DIMENSION_NAMES = (
    'width','height','depth'
)
PIXEL_NAMES = (
    'pixels', 'row', 'column',
)
DATA_SIZE_NAMES = (
    'imageSize',
)

def setImageInput(
    baseOperation, arrayType=None, dimNames=DIMENSION_NAMES,
    pixelName="pixels", typeName=None
):
    """Determine how to convert "pixels" into an image-compatible argument"""
    baseOperation = asWrapper( baseOperation )
    # rank is the count of width,height,depth arguments...
    rank = len([
        # rank is the number of dims we want, not the number we give...
        argName for argName in baseOperation.argNames
        if argName in dimNames
    ]) + 1
    if arrayType:
        converter = TypedImageInputConverter( rank, pixelName, arrayType, typeName=typeName )
        for i,argName in enumerate(baseOperation.argNames):
            if argName in dimNames:
                baseOperation.setPyConverter( argName )
                baseOperation.setCConverter( argName, getattr(converter,argName) )
            elif argName == 'type' and typeName is not None:
                baseOperation.setPyConverter( argName )
                baseOperation.setCConverter( argName, converter.type )
    else:
        converter = ImageInputConverter( rank, pixelsName=pixelName, typeName=typeName or 'type' )
    for argName in baseOperation.argNames:
        if argName in DATA_SIZE_NAMES:
            baseOperation.setPyConverter( argName )
            baseOperation.setCConverter( argName, converter.imageDataSize )
    baseOperation.setPyConverter(
        pixelName, converter,
    )
#	baseOperation.setCResolver(
#		pixelName, converter.cResolver
#	)
    return baseOperation

glDrawPixels = setDimensionsAsInts(
    setImageInput(
        GL_1_1.glDrawPixels
    )
)
glTexSubImage2D = setDimensionsAsInts(
    setImageInput(
        GL_1_1.glTexSubImage2D
    )
)
glTexSubImage1D = setDimensionsAsInts(
    setImageInput(
        GL_1_1.glTexSubImage1D
    )
)
glTexImage2D = setDimensionsAsInts(
    setImageInput(
        GL_1_1.glTexImage2D
    )
)
glTexImage1D = setDimensionsAsInts(
    setImageInput(
        GL_1_1.glTexImage1D
    )
)

def typedImageFunction( suffix, arrayConstant,  baseFunction ):
    """Produce a typed version of the given image function"""
    functionName = baseFunction.__name__
    functionName = '%(functionName)s%(suffix)s'%locals()
    if baseFunction:
        arrayType = arrays.GL_CONSTANT_TO_ARRAY_TYPE[ arrayConstant ]
        function = setDimensionsAsInts(
            setImageInput(
                baseFunction,
                arrayType,
                typeName = arrayConstant,
            )
        )
        return functionName, function
    else:
        return functionName, baseFunction

def _setDataSize( baseFunction, argument='imageSize' ):
    """Set the data-size value to come from the data field"""
    if baseFunction:
        converter = CompressedImageConverter()
        return asWrapper( baseFunction ).setPyConverter(
            argument
        ).setCConverter( argument, converter )
    else:
        return baseFunction

def compressedImageFunction( baseFunction ):
    """Set the imageSize and dimensions-as-ints converters for baseFunction"""
    if baseFunction:
        return setDimensionsAsInts(
            _setDataSize(
                baseFunction, argument='imageSize'
            )
        )
    else:
        return baseFunction

for suffix,arrayConstant in [
    ('b', GL_1_1.GL_BYTE),
    ('f', GL_1_1.GL_FLOAT),
    ('i', GL_1_1.GL_INT),
    ('s', GL_1_1.GL_SHORT),
    ('ub', GL_1_1.GL_UNSIGNED_BYTE),
    ('ui', GL_1_1.GL_UNSIGNED_INT),
    ('us', GL_1_1.GL_UNSIGNED_SHORT),
]:
    for functionName in (
        'glTexImage1D','glTexImage2D',
        'glTexSubImage1D','glTexSubImage2D',
        'glDrawPixels',
        #'glTexSubImage3D','glTexImage3D', # extension/1.2 standard
    ):
        functionName, function = typedImageFunction(
            suffix, arrayConstant, getattr(GL_1_1,functionName),
        )
        globals()[functionName] = function
        try:
            del function, functionName
        except NameError as err:
            pass
    try:
        del suffix,arrayConstant
    except NameError as err:
        pass



from fsui.qt import Qt, QWidget, QHBoxLayout, QPushButton, QApplication
from fsui.qt import QPainter, QRect, QPen, QColor, QPoint, QFont


class TaskBarButton(QWidget):

    def __init__(self, parent, window_handler, window):
        super().__init__(parent)
        self.window = window
        self.window_handler = window_handler
        self.window_handler.window_changed.connect(self.__window_changed)

        self.setFixedHeight(parent.height())
        self.setMaximumWidth(300)
        self.setMinimumWidth(200)

        # self.setAutoFillBackground(True)
        self.setStyleSheet("""
            font-family: "Open Sans";
            font-size: 10pt;
            /* font-weight: bold; */
            /* background-color: #282828; */
            background-color: #000000;
        """)

    def disconnect(self):
        self.window_handler.window_changed.disconnect(self.__window_changed)
        self.window_handler = None
        self.window = None

    def __window_changed(self, window):
        if window != self.window:
            return
        if window.changed == "name":
            self.update()
        elif window.changed == "active":
            self.update()
        elif window.changed == "state":
            self.update()

    def mousePressEvent(self, event):
        if event.button() == 1:
            self.window.activate()
        elif event.button() == 2:
            self.window.open_menu(bottom_left=self.mapToGlobal(QPoint(0, 0)))

    def paintEvent(self, event):
        print("paintEvent")
        painter = QPainter(self)
        # painter.setRenderHints(QPainter.Antialiasing)

        # painter.
        # pen = QPen(QColor(0x0, 0x0, 0x0))
        # pen.setWidth(1)
        # painter.setPen(pen)
        # painter.setBrush(QBrush(Qt.white))
        # rect = QRect(9, 9, 8, 8)

        if self.window.is_active():
            font = painter.font()
            # font.setWeight(QFont.DemiBold)
            font.setWeight(QFont.Bold)
            painter.setFont(font)

        # painter.setBrush(QBrush(QColor(0x50, 0x50, 0x50)))
        painter.setPen(QPen(painter.background().color()))
        painter.setBrush(painter.background())
        painter.drawRect(QRect(0, 0, self.width(), self.height()))

        if self.window.is_minimized():
            painter.setPen(QPen(QColor(0x80, 0x80, 0x80)))
        else:
            painter.setPen(QPen(QColor(0xff, 0xff, 0xff)))
        painter.drawText(QRect(10, 0, self.width(), self.height()),
                         Qt.AlignVCenter, self.window.name)

        painter.setPen(QPen(painter.background().color()))
        # painter.setBrush(painter.background())
        painter.drawRect(QRect(self.width() - 10, self.height(),
                               10, self.height()))

        # painter.setPen(QPen(QColor(0xacccfb)))
        # painter.drawLine(0, 0, self.width(), 0)
        # painter.drawLine(0, 1, 0, self.height())
        #
        # painter.setPen(QPen(QColor(0x5a7cae)))
        # painter.drawLine(self.width() - 1, 0,
        #                  self.width() - 1, self.height())
        # painter.drawLine(0, self.height() - 1,
        #                  self.width(), self.height() - 1)


class TaskBarWidget(QWidget):

    def __init__(self, parent, window_handler):
        super().__init__(parent)
        self.setFixedHeight(28)
        self.window_handler = window_handler

        # self.window_mapping = {}
        self.buttons = []

        # self.setAutoFillBackground(True)
        # palette = self.palette()
        # palette.setColor(self.backgroundRole(), Qt.black)
        # self.setPalette(palette)

        # self.setAutoFillBackground(True)
        self.setStyleSheet("""
            background-color: #000000;
        """)

        self.layout = QHBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(0)

        self.layout.addStretch(0)

        button = QPushButton("X", self)
        button.setFixedHeight(self.height())
        button.setFixedWidth(32)
        button.clicked.connect(self.__close_button_clicked)
        self.layout.addWidget(button)

        self.window_handler.window_added.connect(self.__window_added)
        self.window_handler.window_removed.connect(self.__window_removed)
        # self.window_handler.window_changed.connect(self.__window_changed)

    def __close_button_clicked(self):
        # parent = self
        # while parent.parentWidget():
        #     parent = parent.parentWidget()
        # parent.close()
        QApplication.instance().exit(0)

    def __button_clicked(self, button):
        button.window.activate()

    def __window_added(self, window):
        print("__window_added")
        # button = QPushButton(window.name, self)
        button = TaskBarButton(self, self.window_handler, window)

        button.window = window

        self.layout.insertWidget(len(self.buttons), button)
        self.layout.update()

        self.buttons.append(button)

    def __window_removed(self, window):
        # button = self.window_mapping[window]

        for i, button in enumerate(self.buttons):
            if button.window == window:
                break
        else:
            print("WARNING: did not find button for window", window)
            return

        del self.buttons[i]

        self.layout.removeWidget(button)
        button.disconnect()
        button.setParent(None)
        # button.destroy()
        # del self.window_mapping[window]



import fsui as fsui
from ...Settings import Settings
from ...I18N import gettext


class FullscreenCheckBox(fsui.CheckBox):

    def __init__(self, parent):
        fsui.CheckBox.__init__(self, parent, gettext("Fullscreen"))
        self.on_setting("fullscreen", Settings.get("fullscreen"))
        Settings.add_listener(self)

    def on_destroy(self):
        Settings.remove_listener(self)

    def on_setting(self, key, value):
        if key == "fullscreen":
            self.check(value == "1")

    def on_change(self):
        if self.is_checked():
            Settings.set("fullscreen", "1")
        else:
            Settings.set("fullscreen", "0")



import weakref
from fs_uae_launcher.Config import Config


class AmigaEnableBehavior:

    def __init__(self, parent):
        parent.__amiga_enable_behavior = self
        self._parent = weakref.ref(parent)
        Config.add_listener(self)
        # FIXME
        parent.destroyed.connect(self.on_destroy)

    def on_destroy(self, *args):
        Config.remove_listener(self)

    def on_config(self, key, value):
        if key == "platform":
            widget = self._parent()
            platform = value.lower()
            if platform in ["", "amiga", "cd32", "cdtv"]:
                widget.enable()
            else:
                widget.disable()
